"""Linear solvers.

Documentation of the functionality of this module and examples
can be found at https://bempp.com/handbook/api/solvers.html
"""

from .iterative_solvers import gmres
from .iterative_solvers import cg
from .direct_solvers import lu



"""Iterative solver interfaces."""

import numpy as _np

# pylint: disable=invalid-name
# pylint: disable=too-many-arguments
# pylint: disable=too-many-locals


class IterationCounter(object):
    """Iteration Counter class."""

    def __init__(self, store_residuals, iteration_is_cg=False, operator=None, rhs=None):
        self._count = 0
        self._store_residuals = store_residuals
        self._residuals = []
        self._iteration_is_cg = iteration_is_cg
        self._operator = operator
        self._rhs = rhs

    def __call__(self, x):
        """Call."""
        from bempp_cl.api import log

        self._count += 1
        if self._store_residuals:
            if self._iteration_is_cg:
                res = self._rhs - self._operator * x
            else:
                res = x
            self._residuals.append(_np.linalg.norm(res))
            log(f"GMRES Iteration {self._count} with residual {self._residuals[-1]}")
        else:
            log(f"GMRES Iteration {self._count}")

    @property
    def count(self):
        """Return the number of iterations."""
        return self._count

    @property
    def residuals(self):
        """Return the vector of residuals."""
        return self._residuals


def gmres(
    A,
    b,
    tol=1e-5,
    restart=None,
    maxiter=None,
    use_strong_form=False,
    return_residuals=False,
    return_iteration_count=False,
):
    """Perform GMRES solve via interface to scipy.

    This function behaves like the scipy.sparse.linalg.gmres function. But
    instead of a linear operator and a vector b it takes a boundary operator
    and a grid function or a blocked operator and a list of grid functions.
    The result is returned as a grid function or as a list of grid functions
    in the correct spaces.

    """
    from bempp_cl.api.assembly.boundary_operator import BoundaryOperator
    from bempp_cl.api.assembly.blocked_operator import BlockedOperatorBase

    if isinstance(A, BoundaryOperator):
        return _gmres_single_op_imp(
            A,
            b,
            tol,
            restart,
            maxiter,
            use_strong_form,
            return_residuals,
            return_iteration_count,
        )

    if isinstance(A, BlockedOperatorBase):
        return _gmres_block_op_imp(
            A,
            b,
            tol,
            restart,
            maxiter,
            use_strong_form,
            return_residuals,
            return_iteration_count,
        )

    raise ValueError("A must be a BoundaryOperator or BlockedBoundaryOperator")


def cg(
    A,
    b,
    tol=1e-5,
    maxiter=None,
    use_strong_form=False,
    return_residuals=False,
    return_iteration_count=False,
):
    """Perform CG solve via interface to scipy.

    This function behaves like the scipy.sparse.linalg.cg function. But
    instead of a linear operator and a vector b it takes a boundary operator
    and a grid function. The result is returned as a grid function in the
    correct space.

    """
    from bempp_cl.api.assembly.boundary_operator import BoundaryOperator
    from bempp_cl.api.assembly.grid_function import GridFunction

    import scipy.sparse.linalg

    import bempp_cl.api
    import time

    if not isinstance(A, BoundaryOperator):
        raise ValueError("A must be of type BoundaryOperator")

    if not isinstance(b, GridFunction):
        raise ValueError("b must be of type GridFunction")

    if use_strong_form:
        if not A.range.is_compatible(b.space):
            raise ValueError(
                "The range of A and the domain of A must "
                + "have the same number of unknowns if the strong form is used."
            )
        A_op = A.strong_form()
        b_vec = b.coefficients
    else:
        A_op = A.weak_form()
        b_vec = b.projections(A.dual_to_range)

    callback = IterationCounter(return_residuals, True, A_op, b_vec)
    bempp_cl.api.log("Starting CG iteration")
    start_time = time.time()
    x, info = scipy.sparse.linalg.cg(A_op, b_vec, rtol=tol, maxiter=maxiter, callback=callback)
    end_time = time.time()
    bempp_cl.api.log("CG finished in %i iterations and took %.2E sec." % (callback.count, end_time - start_time))

    res_fun = GridFunction(A.domain, coefficients=x.ravel())

    if return_residuals and return_iteration_count:
        return res_fun, info, callback.residuals, callback.count

    if return_residuals:
        return res_fun, info, callback.residuals

    if return_iteration_count:
        return res_fun, info, callback.count

    return res_fun, info


def _gmres_single_op_imp(
    A,
    b,
    tol=1e-5,
    restart=None,
    maxiter=None,
    use_strong_form=False,
    return_residuals=False,
    return_iteration_count=False,
):
    """Run implementation of GMRES for single operators."""
    from bempp_cl.api.assembly.grid_function import GridFunction

    import scipy.sparse.linalg

    import bempp_cl.api
    import time

    if not isinstance(b, GridFunction):
        raise ValueError("b must be of type GridFunction")

    # Assemble weak form before the logging messages

    if use_strong_form:
        if not A.range.is_compatible(b.space):
            raise ValueError(
                "The range of A and the domain of A must have"
                + "the same number of unknowns if the strong form is used."
            )
        A_op = A.strong_form()
        b_vec = b.coefficients
    else:
        A_op = A.weak_form()
        b_vec = b.projections(A.dual_to_range)

    callback = IterationCounter(return_residuals)

    bempp_cl.api.log("Starting GMRES iteration")
    start_time = time.time()
    x, info = scipy.sparse.linalg.gmres(A_op, b_vec, rtol=tol, restart=restart,
                                        maxiter=maxiter, callback=callback, 
                                        callback_type='legacy')
    end_time = time.time()
    bempp_cl.api.log("GMRES finished in %i iterations and took %.2E sec." % (callback.count, end_time - start_time))

    res_fun = GridFunction(A.domain, coefficients=x.ravel())

    if return_residuals and return_iteration_count:
        return res_fun, info, callback.residuals, callback.count

    if return_residuals:
        return res_fun, info, callback.residuals

    if return_iteration_count:
        return res_fun, info, callback.count

    return res_fun, info


def _gmres_block_op_imp(
    A,
    b,
    tol=1e-5,
    restart=None,
    maxiter=None,
    use_strong_form=False,
    return_residuals=False,
    return_iteration_count=False,
):
    """Run implementation of GMRES for blocked operators."""
    import scipy.sparse.linalg

    import bempp_cl.api
    import time
    from bempp_cl.api.assembly.blocked_operator import (
        coefficients_from_grid_functions_list,
        projections_from_grid_functions_list,
        grid_function_list_from_coefficients,
    )

    # Assemble weak form before the logging messages

    if use_strong_form:
        b_vec = coefficients_from_grid_functions_list(b)
        A_op = A.strong_form()
    else:
        A_op = A.weak_form()
        b_vec = projections_from_grid_functions_list(b, A.dual_to_range_spaces)

    callback = IterationCounter(return_residuals)

    bempp_cl.api.log("Starting GMRES iteration")
    start_time = time.time()
    x, info = scipy.sparse.linalg.gmres(A_op, b_vec, rtol=tol, restart=restart, 
                                        maxiter=maxiter, callback=callback,
                                        callback_type='legacy')
    end_time = time.time()
    bempp_cl.api.log("GMRES finished in %i iterations and took %.2E sec." % (callback.count, end_time - start_time))

    res_fun = grid_function_list_from_coefficients(x.ravel(), A.domain_spaces)

    if return_residuals and return_iteration_count:
        return res_fun, info, callback.residuals, callback.count

    if return_residuals:
        return res_fun, info, callback.residuals

    if return_iteration_count:
        return res_fun, info, callback.count

    return res_fun, info



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct  3 15:56:03 2023

@author: tom.munoz
"""

from electroacPy.global_ import air
from electroacPy.speakerSim.electroAcousticDriver import electroAcousticDriver as ead
from electroacPy.general.freqop import freq_log10, laplace
import matplotlib.pyplot as plt
import numpy as np

# circuit solver
import electroacPy.circuitSolver.components.electric as compe
import electroacPy.circuitSolver.components.acoustic as compa
from electroacPy.circuitSolver.blocks.electrodynamic import EAD
from electroacPy.circuitSolver.solver import circuit

pi = np.pi

## New approach
class speakerBox:
    def __init__(self, Vb, frequencyRange=freq_log10(20, 2500, 50),
                 Qab=120, Qal=30, c=air.c, rho=air.rho, **kwargs):
        """
        Setup a louspeaker enclosure as a lumped-element object.

        :param Vb: float,
            Back volume (set behind the drive unit)
        :param frequencyRange: array,

        :param Qab: float,
            Quality factor linked to losses in enclosure (damping). Default is 120
        :param Qal: float,
            Quality factor linked to losses out of enclosure (leakage). Default is 30
        :param c: float,
            Speed of sound in air. Default is 343 m/s
        :param rho: float,
            Air density. Default is 1.22 kg/m^3
        """
        # simulation parameters
        self.c = c
        self.rho = rho
        self.Qab = Qab
        self.Qal = Qal        
        self.frequencyRange = frequencyRange
        self.s = laplace(frequencyRange)


        # mandatory parameter
        # 0. Sealed enclosure
        self.Vb = Vb
        self.config = "sealed"
        if kwargs:
            self.detectConfig(**kwargs)

        ## PROCESS
        # self.computeImpedance()

        ## TO STORE RESULTS
        # velocity
        self.v    = np.zeros(len(frequencyRange), dtype=complex)
        self.vp   = np.zeros(len(frequencyRange), dtype=complex)
        self.vp2  = np.zeros(len(frequencyRange), dtype=complex)
        self.vpr  = np.zeros(len(frequencyRange), dtype=complex)
        self.vpr2 = np.zeros(len(frequencyRange), dtype=complex)
        # volume velocity
        self.Q    = np.zeros(len(frequencyRange), dtype=complex)
        self.Qp   = np.zeros(len(frequencyRange), dtype=complex)
        self.Qp2  = np.zeros(len(frequencyRange), dtype=complex)
        self.Qpr  = np.zeros(len(frequencyRange), dtype=complex)
        self.Qpr2 = np.zeros(len(frequencyRange), dtype=complex)
        # impedance
        self.Ze = np.zeros(len(frequencyRange), dtype=complex)  # total electrical impedance of driver in enclosure

        # reference to driver
        self.whichDriver = None
        self.Nd = 1
        self.wiring = "parallel"

        # acoustic simulation reference
        self.isFEM = False
        self.isBEM = False
        self.ref2bem = None  # is it referenced to bem mesh ?
        self.poly_data = False  # is class from polytech?


    def detectConfig(self, **kwargs):
        """
        Set enclosure config depending on input arguments.
        :param kwargs:
        :return:
        """
        ## 6th order bandpass with ports
        if ("Lp" in kwargs and ("Sp" in kwargs or "rp" in kwargs) and
                "Vf" in kwargs and "Lp2" in kwargs and ("Sp2" in kwargs or "rp2" in kwargs)):
            self.config = "bandpass_2"
            self.flange = "single"
            for key, value in kwargs.items():
                setattr(self, key, value)
                if key == "Sp":
                    self.rp = np.sqrt(value / np.pi)
                elif key == "rp":
                    self.Sp = np.pi * value**2
                elif key == "Sp2":
                    self.rp2 = np.sqrt(value / np.pi)
                elif key == "rp2":
                    self.Sp2 = np.pi * value**2
                elif key == "flange":
                    self.flange = value

        ## 6th order bandpass with passive radiators
        elif ("Mmd" in kwargs and "Cmd" in kwargs and "Rmd" in kwargs and
              "Vf" in kwargs and "Mmd2" in kwargs and "Cmd2" in kwargs and "Rmd2" in kwargs and
              "Sd" in kwargs and "Sd2" in kwargs):
            self.config = "bandpass_pr_2"
            for key, value in kwargs.items():
                setattr(self, key, value)

        ## 4th order bandpass with port
        elif "Lp" in kwargs and ("Sp" in kwargs or "rp" in kwargs) and "Vf" in kwargs:
            self.config = "bandpass"
            self.flange = "single"
            for key, value in kwargs.items():
                setattr(self, key, value)
                if key == "Sp":
                    self.rp = np.sqrt(value / np.pi)
                elif key == "rp":
                    self.Sp = np.pi * value**2
                elif key == "flange":
                    self.flange = value

        ## 4th order bandpass with passive radiator
        elif ("Mmd" in kwargs and "Cmd" in kwargs and
              "Rmd" in kwargs and "Vf" in kwargs and "Sd" in kwargs):
            self.config = "bandpass_pr"
            for key, value in kwargs.items():
                setattr(self, key, value)

        ## ported enclosure
        elif "Lp" in kwargs and ("Sp" in kwargs or "rp" in kwargs):
            self.config = "vented"
            self.flange = "single"
            for key, value in kwargs.items():
                setattr(self, key, value)
                if key == "Sp":
                    self.rp = np.sqrt(value / np.pi)
                elif key == "rp":
                    self.Sp = np.pi * value**2
                elif key == "flange":
                    self.flange = value

        ## passive radiator enclosure
        elif "Mmd" in kwargs and "Cmd" in kwargs and "Rmd" in kwargs and "Sd" in kwargs:
            self.config = "passiveRadiator"
            for key, value in kwargs.items():
                setattr(self, key, value)
        else:
            raise ValueError("Invalid configuration: The provided kwargs do not match any valid configuration.\n"
                             "1. Ported enclosure: Lp and (Sp or rp), \n"
                             "2. Passive radiator: Mmd, Cmd and Rmd, \n"
                             "3. 4th order bandpass (port): Vf, Lp and (Sp or rp), \n"
                             "4. 4th order bandpass (passive radiator): Vf, Mmd, Cmd and Rmd, \n"
                             "5. 6th order bandpass (port): Vf, Lp, (Sp or rp), Lp2 and (Sp2 or rp2), \n"
                             "6. 6th order bandpass (passive radiator): Vf, Mmd, Cmd, Rmd, Mmd2, Cmd2 and Rmd2.")
        return None

    ### =============================
    ## ACOUSTICAL IMPEDANCE FUNCTIONS
    def computeImpedance(self):
        """
        Just a small function to send impedance calc to proper function
        :return: None
        """

        if self.config == "sealed":
            self.Za_in = self.sealed_box()

        elif self.config == "vented":
            self.Za_in, self.Zab, self.Zap = self.vented_box()

        elif self.config == "passiveRadiator":
            self.Za_in, self.Zab, self.Zapr = self.pr_box()

        elif self.config == "bandpass":
            self.Za_in, self.Zab, self.Zaf, self.Zap = self.bp4_box()

        elif self.config == "bandpass_2":
            self.Za_in, self.Zab, self.Zaf, self.Zap, self.Zap2 = self.bp6_box()

        elif self.config == "bandpass_pr":
            self.Za_in, self.Zab, self.Zaf, self.Zapr = self.bp4_pr_box()
        return None

    ### ======================================
    ## IMPEDANCE AND VOLUME VELOCITY FUNCTIONS
    def sealed_box(self, driver):   
        # prepare some values for components
        Cab = self.Vb / self.rho / self.c**2    # box volume
        wc  = 2 * np.pi * driver.Fs * np.sqrt(1 + driver.Vas / self.Vb)
        Rab = 1 / wc / Cab / self.Qab    # internal losses  
        Ral = self.Qal / wc / Cab        # leakage 
        
        # define components
        enclosure = circuit(self.frequencyRange)
        U   = compe.voltageSource(1, 0, driver.U)
        DRV = EAD(1, 0, 2, 3, driver.Le, driver.Re, 
                  driver.Cms, driver.Mms, driver.Rms, 
                  driver.Bl, driver.Sd, v_probe="v") 
        RAL = compe.resistance(3, 0, Ral)
        RAB = compe.resistance(3, 4, Rab)
        CAB = compe.capacitance(4, 0, Cab)
        RAD = compa.radiator(2, 0, driver.Sd, self.rho, self.c)
        
        # setup and run
        enclosure.addComponent(U, RAL, RAB, CAB, RAD)
        enclosure.addBlock(DRV)
        enclosure.run()
        
        # extract data
        Q  = enclosure.getPotential(2) * RAD.Gs
        v  = enclosure.getFlow("v")
        Ze = -enclosure.getPotential(1) / enclosure.getFlow(1)
        return Q, v, Ze

    def vented_box(self, driver):  
        # prepare some values for components
        Cab = self.Vb / self.rho / self.c**2    # box volume
        wc  = 2 * np.pi * driver.Fs * np.sqrt(1 + driver.Vas / self.Vb)
        Rab = 1 / wc / Cab / self.Qab    # internal losses  
        Ral = self.Qal / wc / Cab        # leakage
        
        # define component
        enclosure = circuit(self.frequencyRange)
        U    = compe.voltageSource(1, 0, driver.U)
        DRV  = EAD(1, 0, 2, 3, driver.Le, driver.Re, 
                   driver.Cms, driver.Mms, driver.Rms, 
                   driver.Bl, driver.Sd, v_probe="v") 
        
        RAL = compe.resistance(3, 0, Ral)
        RAB = compe.resistance(3, 4, Rab)
        CAB = compe.capacitance(4, 0, Cab)

        RAD  = compa.radiator(2, 0, driver.Sd, self.rho, self.c)
        PORT = compa.port(3, 5, self.Lp, self.rp, 
                          self.flange, rho=self.rho, c=self.c)
        RADP = compa.radiator(5, 0, self.Sp, self.rho, self.c)
        
        # setup and run
        enclosure.addComponent(U, RAL, RAB, CAB, RAD, PORT, RADP)
        enclosure.addBlock(DRV)
        enclosure.run()
        
        # extract data
        Q  = enclosure.getPotential(2) * RAD.Gs
        Qp = enclosure.getPotential(5) * RADP.Gs
        v  = enclosure.getFlow("v")
        vp =  Qp / self.Sp
        Ze = -enclosure.getPotential(1) / enclosure.getFlow(1)
        return Q, Qp, v, vp, Ze

    def passive_radiator(self, driver):
        # prepare some values for components
        Cab = self.Vb / self.rho / self.c**2    # box volume
        wc  = 2 * np.pi * driver.Fs * np.sqrt(1 + driver.Vas / self.Vb)
        Rab = 1 / wc / Cab / self.Qab    # internal losses  
        Ral = self.Qal / wc / Cab 
        
        # define components
        enclosure = circuit(self.frequencyRange)
        U     = compe.voltageSource(1, 0, driver.U)
        DRV   = EAD(1, 0, 2, 3, driver.Le, driver.Re, 
                    driver.Cms, driver.Mms, driver.Rms, 
                    driver.Bl, driver.Sd, v_probe="v") 

        # box
        RAL = compe.resistance(3, 0, Ral)        
        RAB = compe.resistance(3, 4, Rab)
        CAB = compe.capacitance(4, 0, Cab)
        
        RAD   = compa.radiator(2, 0, driver.Sd, self.rho, self.c)
        PR    = compa.membrane(3, 5, self.Cmd, self.Mmd, self.Rmd, self.Sd, self.rho, self.c)
        RADPR = compa.radiator(5, 0, self.Sd, self.rho, self.c)       
        
        # setup and run
        enclosure.addComponent(U, RAL, RAB, CAB, RAD, PR, RADPR)
        enclosure.addBlock(DRV)
        enclosure.run()
        
        # extract data
        Q   = enclosure.getPotential(2) * RAD.Gs
        Qpr = enclosure.getPotential(5) * RADPR.Gs
        v   = enclosure.getFlow("v")
        vpr =  Qpr / self.Sd
        Ze  = -enclosure.getPotential(1) / enclosure.getFlow(1)
        return Q, Qpr, v, vpr, Ze

    def bandpass4_port(self, driver):
        # for simplicity, we consider same losses for front and back enclosure
        # prepare some values for components
        Cab = self.Vb / self.rho / self.c**2    # box volume
        wc  = 2 * np.pi * driver.Fs * np.sqrt(1 + driver.Vas / self.Vb)
        Rab = 1 / wc / Cab / self.Qab    # internal losses  
        Ral = self.Qal / wc / Cab        # leakage 
        
        Cabf = self.Vf / self.rho / self.c**2
        wcf  = 2 * np.pi * driver.Fs * np.sqrt(1 + driver.Vas / self.Vf)
        Rabf = 1 / wcf / Cabf / self.Qab    # internal losses (front)
        Ralf = self.Qal / wcf / Cabf        # leakage (front)

        # define component
        enclosure = circuit(self.frequencyRange)
        U     = compe.voltageSource(1, 0, driver.U)
        DRV   = EAD(1, 0, 2, 5, driver.Le, driver.Re, 
                    driver.Cms, driver.Mms, driver.Rms, 
                    driver.Bl, driver.Sd, v_probe="v") 
    
        RALF = compe.resistance(2, 0, Ralf)
        RABF = compe.resistance(2, 3, Rabf)
        CABF = compe.capacitance(3, 0, Cabf)
        PORTF = compa.port(2, 4, self.Lp, self.rp, 
                           self.flange, rho=self.rho, c=self.c)
        RADPF = compa.radiator(4, 0, self.Sp, self.rho, self.c)
        
        RAL = compe.resistance(5, 0, Ral)        
        RAB = compe.resistance(5, 6, Rab)
        CAB = compe.capacitance(6, 0, Cab)    
    
        # setup and run
        enclosure.addComponent(U, RALF, RABF, CABF, 
                       RAL, RAB, CAB, PORTF, RADPF)
        enclosure.addBlock(DRV)
        enclosure.run()
        
        # extract data
        Qp = enclosure.getPotential(4) * RADPF.Gs
        vp = Qp / self.Sp
        v  = enclosure.getFlow("v")
        Ze = -enclosure.getPotential(1) / enclosure.getFlow(1)
        return Qp, vp, v, Ze

    def bandpass6_port(self, driver):
        # for simplicity, we consider same losses for front and back enclosure
        # prepare some values for components
        Cab = self.Vb / self.rho / self.c**2    # box volume
        wc  = 2 * np.pi * driver.Fs * np.sqrt(1 + driver.Vas / self.Vb)
        Rab = 1 / wc / Cab / self.Qab    # internal losses  
        Ral = self.Qal / wc / Cab        # leakage 
        
        Cabf = self.Vf / self.rho / self.c**2
        wcf  = 2 * np.pi * driver.Fs * np.sqrt(1 + driver.Vas / self.Vf)
        Rabf = 1 / wcf / Cabf / self.Qab    # internal losses (front)
        Ralf = self.Qal / wcf / Cabf        # leakage (front)
        
        # define component
        enclosure = circuit(self.frequencyRange)
        U     = compe.voltageSource(1, 0, driver.U)
        DRV   = EAD(1, 0, 2, 5, driver.Le, driver.Re, 
                    driver.Cms, driver.Mms, driver.Rms, 
                    driver.Bl, driver.Sd, v_probe="v") 

        RALF  = compe.resistance(2, 0, Ralf)
        RABF  = compe.resistance(2, 3, Rabf)
        CABF  = compe.capacitance(3, 0, Cabf)
        PORTF = compa.port(2, 4, self.Lp, self.rp, 
                           self.flange, rho=self.rho, c=self.c)
        RADPF = compa.radiator(4, 0, self.Sp, self.rho, self.c)
        
        RAL   = compe.resistance(5, 0, Ral)
        RAB   = compe.resistance(5, 6, Rab)
        CAB   = compe.capacitance(6, 0, Cab)
        PORTB = compa.port(5, 7, self.Lp2, self.rp2, 
                           self.flange, rho=self.rho, c=self.c)
        RADPB = compa.radiator(7, 0, self.Sp2, self.rho, self.c)
        
        # setup and run
        enclosure.addComponent(U, RALF, RABF, CABF,
                       RAL, RAB, CAB, PORTF, RADPF, 
                       PORTB, RADPB)
        enclosure.addBlock(DRV)
        enclosure.run()
        

        # extract data
        Qp = enclosure.getPotential(4) * RADPF.Gs  # front 
        vp = Qp / self.Sp
        Qp2 = enclosure.getPotential(7) * RADPB.Gs # back
        vp2 = Qp2 / self.Sp2
        v  = enclosure.getFlow("v")
        Ze = -enclosure.getPotential(1) / enclosure.getFlow(1)
        return Qp, vp, Qp2, vp2, v, Ze


    def bandpass4_passive_radiator(self, driver):
        # for simplicity, we consider same losses for front and back enclosure
        # prepare some values for components
        Cab = self.Vb / self.rho / self.c**2    # box volume
        wc  = 2 * np.pi * driver.Fs * np.sqrt(1 + driver.Vas / self.Vb)
        Rab = 1 / wc / Cab / self.Qab    # internal losses  
        Ral = self.Qal / wc / Cab        # leakage 
        
        Cabf = self.Vf / self.rho / self.c**2
        wcf  = 2 * np.pi * driver.Fs * np.sqrt(1 + driver.Vas / self.Vf)
        Rabf = 1 / wcf / Cabf / self.Qab    # internal losses (front)
        Ralf = self.Qal / wcf / Cabf        # leakage (front)
        
        # define component
        enclosure = circuit(self.frequencyRange)
        U     = compe.voltageSource(1, 0, driver.U)
        DRV   = EAD(1, 0, 2, 5, driver.Le, driver.Re, 
                    driver.Cms, driver.Mms, driver.Rms, 
                    driver.Bl, driver.Sd, v_probe="v") 

        RALF  = compe.resistance(2, 0, Ralf)
        RABF  = compe.resistance(2, 3, Rabf)
        CABF  = compe.capacitance(3, 0, Cabf)
        PRF   = compa.membrane(2, 4, self.Cmd, self.Mmd, self.Rmd, self.Sd, self.rho, self.c)
        RADPF = compa.radiator(4, 0, self.Sd, self.rho, self.c)
        
        RAL = compe.resistance(5, 0, Ral)        
        RAB = compe.resistance(5, 6, Rab)
        CAB = compe.capacitance(6, 0, Cab)
    
        # setup and run
        enclosure.addComponent(U, RALF, RABF, CABF, PRF, RADPF,
                               RAL, RAB, CAB)
        enclosure.addBlock(DRV)
        enclosure.run()
        
        # extract data
        Qpr = enclosure.getPotential(4) * RADPF.Gs
        vpr = Qpr / self.Sd
        v   = enclosure.getFlow("v")
        Ze  = -enclosure.getPotential(1) / enclosure.getFlow(1)
        return Qpr, vpr, v, Ze
    
    
    def bandpass6_passive_radiator(self, driver):     
        # for simplicity, we consider same losses for front and back enclosure
        # prepare some values for components
        Cab = self.Vb / self.rho / self.c**2    # box volume
        wc  = 2 * np.pi * driver.Fs * np.sqrt(1 + driver.Vas / self.Vb)
        Rab = 1 / wc / Cab / self.Qab    # internal losses  
        Ral = self.Qal / wc / Cab        # leakage 
        
        Cabf = self.Vf / self.rho / self.c**2
        wcf  = 2 * np.pi * driver.Fs * np.sqrt(1 + driver.Vas / self.Vf)
        Rabf = 1 / wcf / Cabf / self.Qab    # internal losses (front)
        Ralf = self.Qal / wcf / Cabf        # leakage (front)
        
        # define component
        enclosure = circuit(self.frequencyRange)
        U     = compe.voltageSource(1, 0, driver.U)
        DRV   = EAD(1, 0, 2, 4, driver.Le, driver.Re, 
                    driver.Cms, driver.Mms, driver.Rms, 
                    driver.Bl, driver.Sd, v_probe="v") 
        
        RALF  = compe.resistance(2, 0, Ralf)
        RABF  = compe.resistance(2, 3, Rabf)
        CABF  = compe.capacitance(3, 0, Cabf)
        PRF   = compa.membrane(2, 4, self.Cmd, self.Mmd, self.Rmd, self.Sd, self.rho, self.c)
        RADPF = compa.radiator(4, 0, self.Sd, self.rho, self.c)
        
        RAL   = compe.resistance(5, 0, Ral)
        RAB   = compe.resistance(5, 6, Rab)
        CAB   = compe.capacitance(6, 0, Cab)
        PRB   = compa.membrane(5, 7, self.Cmd2, self.Mmd2, self.Rmd2, self.Sd2, self.rho, self.c) 
        RADPB = compa.radiator(7, 0, self.Sd2, self.rho, self.c)
        
        
        # setup and run
        # enclosure.addComponent(U, BOXF, PRF, RADPF, BOXB, PRB, RADPB)
        enclosure.addComponent(U, RALF, RABF, CABF, PRF, RADPF,
                       RAL, RAB, CAB, PRB, RADPB)
        enclosure.addBlock(DRV)
        enclosure.run()
        
        # extract data
        Qpr  = enclosure.getPotential(4) * RADPF.Gs
        vpr  = Qpr / self.Sd
        Qpr2 = enclosure.getPotential(7) * RADPB.Gs
        vpr2 = Qpr2 / self.Sd2
        v    = enclosure.getFlow("v")
        Ze   = -enclosure.getPotential(1) / enclosure.getFlow(1)
        return Qpr, vpr, Qpr2, vpr2, v, Ze


    ## SET DRIVERS IN ENCLOSURE
    def getDriverResponse(self, driver, Nd=1, wiring='parallel'):
        d = driver
        n = Nd
        self.Nd = Nd
        self.wiring = wiring

        if wiring == 'parallel':
            drvTmp = ead(d.U, d.Le / n, d.Re / n, d.Cms / n, d.Mms * n, d.Rms * n, d.Bl, d.Sd * n, d.f_array, d.c,
                         d.rho)
        elif wiring == 'series':
            drvTmp = ead(d.U, d.Le * n, d.Re * n, d.Cms / n, d.Mms * n, d.Rms * n, d.Bl * n, d.Sd * n, d.f_array, d.c,
                         d.rho)
        else:
            raise ValueError("Speaker wiring not understood. Accepted values are: 'parallel', 'series'.")

        ## Compute volume / particule velocity -> add enclosure config here
        if self.config == "sealed":
            self.Q, self.v, self.Ze = self.sealed_box(drvTmp)
        elif self.config == "vented":
            self.Q, self.Qp, self.v, self.vp, self.Ze = self.vented_box(drvTmp)
        elif self.config == "passiveRadiator":
            self.Q, self.Qpr, self.v, self.vpr, self.Ze = self.passive_radiator(drvTmp)
        elif self.config == "bandpass":
            self.Qp, self.vp, self.v, self.Ze = self.bandpass4_port(drvTmp)
        elif self.config == "bandpass_2":
            self.Qp, self.vp, self.Qp2, self.vp2, self.v, self.Ze = self.bandpass6_port(drvTmp)
        elif self.config == "bandpass_pr":
            self.Qpr, self.vpr, self.v, self.Ze = self.bandpass4_passive_radiator(drvTmp)
        elif self.config == "bandpass_pr_2":
            self.Qpr, self.vpr, self.Qpr2, self.vpr2, self.v, self.Ze = self.bandpass6_passive_radiator(drvTmp)
        return None


    def plotZe(self, **kwargs):
        """
        Plot the electrical impedance ZeTot in both modulus and phase.
    
        Returns
        -------
        None
    
        """
        
        if "figsize" in kwargs:
            size=kwargs["figsize"]
        else:
            size=None
        
        fig, ax = plt.subplots(2, 1, figsize=size)
        ax[0].semilogx(self.frequencyRange, np.abs(self.Ze))
        ax[0].set(ylabel="Magnitude [Ohm]")
        
        ax[1].semilogx(self.frequencyRange, np.angle(self.Ze))
        ax[1].set(xlabel="Frequency [Hz]", ylabel="Phase [rad]")
        for i in range(2):
            ax[i].grid(which="both", linestyle="dotted")
        plt.tight_layout()
        
        if "savefig" in kwargs:
            path = kwargs["savefig"]
            plt.savefig(path)
      
        return plt.show()
    
    def plotXVA(self, **kwargs):
        """
        Plot the displacement, velocity, and acceleration frequency responses.
    
        Returns
        -------
        None
    
        """
        
        if "figsize" in kwargs:
            size=kwargs["figsize"]
        else:
            size=None
        
        
        # convert velocity into displacement and acceleration
        x = self.v / laplace(self.frequencyRange)
        a = self.v * laplace(self.frequencyRange)
        
        xp   = self.vp / laplace(self.frequencyRange)
        ap   = self.vp * laplace(self.frequencyRange)
        xp2  = self.vp2 / laplace(self.frequencyRange)
        ap2  = self.vp2 * laplace(self.frequencyRange)
        xpr  = self.vpr / laplace(self.frequencyRange)
        apr  = self.vpr * laplace(self.frequencyRange)
        xpr2 = self.vpr2 / laplace(self.frequencyRange)
        apr2 = self.vpr2 * laplace(self.frequencyRange)
        
        fig, ax = plt.subplots(3, 1, figsize=size)
        ax[0].semilogx(self.frequencyRange, np.abs(x), label='Displacement')
        ax[1].semilogx(self.frequencyRange, np.abs(self.v), label='Velocity')
        ax[2].semilogx(self.frequencyRange, np.abs(a), label='Acceleration')
        ax[2].set(xlabel="Frequency [Hz]")
        ax[0].set(ylabel="mm", title="Driver")
        ax[1].set(ylabel="m/s")
        ax[2].set(ylabel="m/s^2", xlabel="Frequency [Hz]")
        for i in range(3):
            ax[i].grid(which='both', linestyle="dotted")
            ax[i].legend(loc='best')
        plt.tight_layout()
        
        if "savefig" in kwargs:
            path = kwargs["savefig"]
            plt.savefig(path)
      
        
        if np.all(self.vp != 0):
            fig, ax = plt.subplots(3, 1, figsize=size)
            ax[0].semilogx(self.frequencyRange, np.abs(xp), label='Displacement')
            ax[1].semilogx(self.frequencyRange, np.abs(self.vp), label='Velocity')
            ax[2].semilogx(self.frequencyRange, np.abs(ap), label='Acceleration')
            ax[0].set(ylabel="mm", title="Port")
            ax[1].set(ylabel="m/s")
            ax[2].set(ylabel="m/s^2", xlabel="Frequency [Hz]")
            for i in range(3):
                ax[i].grid(which='both', linestyle="dotted")
                ax[i].legend(loc='best')
            plt.tight_layout()
            if "savefig" in kwargs:
                path = kwargs["savefig"]
                extension = path[-4:]
                plt.savefig(path[:-4] + "_vp" + extension)       
            
        if np.all(self.vp2 != 0):
            fig, ax = plt.subplots(3, 1, figsize=size)
            ax[0].semilogx(self.frequencyRange, np.abs(xp2), label='Displacement')
            ax[1].semilogx(self.frequencyRange, np.abs(self.vp2), label='Velocity')
            ax[2].semilogx(self.frequencyRange, np.abs(ap2), label='Acceleration')
            ax[0].set(ylabel="mm", title="Port - 2")
            ax[1].set(ylabel="m/s")
            ax[2].set(ylabel="m/s^2", xlabel="Frequency [Hz]")
            for i in range(3):
                ax[i].grid(which='both', linestyle="dotted")
                ax[i].legend(loc='best')
            plt.tight_layout()
            if "savefig" in kwargs:
                path = kwargs["savefig"]
                extension = path[-4:]
                plt.savefig(path[:-4] + "_vp2" + extension)
            
        if np.all(self.vpr != 0):
            fig, ax = plt.subplots(3, 1, figsize=size)
            ax[0].semilogx(self.frequencyRange, np.abs(xpr), label='Displacement')
            ax[1].semilogx(self.frequencyRange, np.abs(self.vpr), label='Velocity')
            ax[2].semilogx(self.frequencyRange, np.abs(apr), label='Acceleration')
            ax[0].set(ylabel="mm", title="Passive-radiator")
            ax[1].set(ylabel="m/s")
            ax[2].set(ylabel="m/s^2", xlabel="Frequency [Hz]")
            for i in range(3):
                ax[i].grid(which='both', linestyle="dotted")
                ax[i].legend(loc='best')  
            plt.tight_layout()
            if "savefig" in kwargs:
                path = kwargs["savefig"]
                extension = path[-4:]
                plt.savefig(path[:-4] + "_vpr" + extension)
            
        if np.all(self.vpr2 != 0):
            fig, ax = plt.subplots(3, 1, figsize=size)
            ax[0].semilogx(self.frequencyRange, np.abs(xpr2), label='Displacement')
            ax[1].semilogx(self.frequencyRange, np.abs(self.vpr2), label='Velocity')
            ax[2].semilogx(self.frequencyRange, np.abs(apr2), label='Acceleration')
            ax[0].set(ylabel="mm", title="Passive-radiator - 2")
            ax[1].set(ylabel="m/s")
            ax[2].set(ylabel="m/s^2", xlabel="Frequency [Hz]")
            for i in range(3):
                ax[i].grid(which='both', linestyle="dotted")
                ax[i].legend(loc='best')   
            plt.tight_layout()
            if "savefig" in kwargs:
                path = kwargs["savefig"]
                extension = path[-4:]
                plt.savefig(path[:-4] + "_vpr2" + extension)
        
        return plt.show()
    
    def exportZe(self, folder_name, file_name):
        import os
        
        # Create the folder if it doesn't exist
        if not os.path.exists(folder_name):
            os.makedirs(folder_name)
        
        module = np.abs(self.Ze)
        phase = np.angle(self.Ze, deg=True)
        path = os.path.join(folder_name, file_name)
        np.savetxt(path, np.array([self.frequencyRange, module, phase]).T,
                   fmt="%.3f",
                   header="Freq[Hz]  Imp[Ohm]  Phase[Deg]",
                   delimiter=',',
                   comments='')
        
        
        
        
        
        



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct  3 15:55:45 2023

@author: tom.munoz
"""

import numpy as np
import matplotlib.pyplot as plt
import electroacPy.general as gtb
from scipy.signal import butter, freqs, freqz
from numpy import sin, cos, sqrt, pi
import scipy as sp

class xovers:
    """
    A class for designing analog crossover filters using biquad filter stages.

    This class provides methods to add various types of analog crossover filters to a design,
    including low-pass, high-pass, band-pass, peak EQ, low-shelf, high-shelf, gain, and delay filters.
    It also supports deleting filters, updating the frequency response, and plotting the overall response.

    Attributes:
        f_axis (array-like): The frequency axis for which the filters are designed.
        h (array-like, dtype=complex): The current total frequency response of the crossover.
        filterList (list): List of filter names in the crossover.
        frf_List (list): List of individual filter frequency responses.

    Methods:
        addLowPass(filterName, order, fc): Add an analog low-pass filter.
        addHighPass(filterName, order, fc): Add an analog high-pass filter.
        addBandPass(filterName, order, fc1, fc2): Add an analog band-pass filter.
        addLowPassBQ(filterName, fc, Q, dBGain=0, coeff=False): Add a biquad low-pass filter.
        addHighPassBQ(filterName, fc, Q, dBGain=0, coeff=False): Add a biquad high-pass filter.
        addBandPassBQ(filterName, fc, Q, dBGain=0, coeff=False): Add a biquad band-pass filter.
        addPeakEQ(filterName, fc, Q, dBGain=0, coeff=False): Add a biquad peak equalizer filter.
        addLowShelf(filterName, fc, Q, dBGain=0, coeff=False): Add a biquad low-shelf filter.
        addHighShelf(filterName, fc, Q, dBGain=0, coeff=False): Add a biquad high-shelf filter.
        addGain(filterName, dBGain): Add a gain filter.
        addDelay(filterName, dt, timeConvention='-jwt'): Add a time delay filter.
        deleteFilter(filterName): Delete a filter from the crossover.
        plot(divide_stages=True, amplitude=40): Plot the overall frequency response.
    """
    def __init__(self, f_axis):
        self.identifier = "XO"
        self.f_axis = f_axis
        self.h = np.ones(len(f_axis), dtype=complex)
        self.filterList = []
        self.frf_List = []

        # ref 2 study
        self.referenceStudy = None
        self.ref2bem = None

    def addLowPass(self, filterName, order, fc):
        """
        Add an analog low-pass filter to the crossover design.

        Parameters:
            filterName (str): Name of the filter to be added.
            order (int): Order of the low-pass filter (1 or higher).
            fc (float): Cutoff frequency of the filter in Hertz.

        Returns:
            None

        Notes:
            This method adds an analog low-pass filter to the crossover design using the given
            order and cutoff frequency. The filter coefficients are computed using the `butter`
            function from the SciPy library. The frequency response is calculated using the `freqs`
            function.

            After adding the filter, the crossover's `filterList` and `frf_List` attributes are updated,
            and the total frequency response `h` is recalculated using the `updateCrossovers` function.
        """
        self.checkExistingFilters(filterName)

        b, a = butter(order, fc, btype='low', analog=True)
        _, h = freqs(b, a, self.f_axis)

        # update crossover
        self.filterList.append(filterName)
        self.frf_List.append(h)
        self.h = updateCrossovers(self.f_axis, self.filterList, self.frf_List)
        return None

    def addHighPass(self, filterName, order, fc):
        """
        Add an analog high-pass filter to the crossover design.

        Parameters:
            filterName (str): Name of the filter to be added.
            order (int): Order of the high-pass filter (1 or higher).
            fc (float): Cutoff frequency of the filter in Hertz.

        Returns:
            None

        Notes:
            This method adds an analog high-pass filter to the crossover design using the given
            order and cutoff frequency. The filter coefficients are computed using the `butter`
            function from the SciPy library. The frequency response is calculated using the `freqs`
            function.

            After adding the filter, the crossover's `filterList` and `frf_List` attributes are updated,
            and the total frequency response `h` is recalculated using the `updateCrossovers` function.
        """
        self.checkExistingFilters(filterName)

        b, a = butter(order, fc, btype='high', analog=True)
        _, h = freqs(b, a, self.f_axis)

        # update crossover
        self.filterList.append(filterName)
        self.frf_List.append(h)
        self.h = updateCrossovers(self.f_axis, self.filterList, self.frf_List)
        return None

    def addBandPass(self, filterName, order, fc1, fc2):
        """
        Add an analog band-pass filter to the crossover design.

        Parameters:
            filterName (str): Name of the filter to be added.
            order (int): Order of the band-pass filter (1 or higher).
            fc1 (float): Lower cutoff frequency of the filter in Hertz.
            fc2 (float): Upper cutoff frequency of the filter in Hertz.

        Returns:
            None

        Notes:
            This method adds an analog band-pass filter to the crossover design using the given
            order, lower cutoff frequency (fc1), and upper cutoff frequency (fc2). The filter
            coefficients are computed using the `butter` function from the SciPy library. The
            frequency response is calculated using the `freqs` function.

            After adding the filter, the crossover's `filterList` and `frf_List` attributes are updated,
            and the total frequency response `h` is recalculated using the `updateCrossovers` function.

        """
        # analog band-pass filter
        self.checkExistingFilters(filterName)

        b, a = butter(order, [fc1, fc2], btype='bandpass', analog=True)
        _, h = freqs(b, a, self.f_axis)

        # update crossover
        self.filterList.append(filterName)
        self.frf_List.append(h)
        self.h = updateCrossovers(self.f_axis, self.filterList, self.frf_List)
        return None

    def addLowPassBQ(self, filterName, fc, Q, dBGain=0, coeff=False):
        """
        Add a biquad (BQ) low-pass filter to the crossover design.

        Parameters:
            filterName (str): Name of the filter to be added.
            fc (float): Cutoff frequency of the filter in Hertz.
            Q (float): Quality factor of the filter.
            dBGain (float): Gain in decibels applied to the filter (default: 0).
            coeff (bool): If True, return filter coefficients (default: False).

        Returns:
            list or None: If coeff is True, returns a list containing filter coefficients [b, a].
                          Otherwise, returns None.

        Notes:
            This method adds a biquad low-pass filter to the crossover design using the given
            cutoff frequency (fc), quality factor (Q), and optional gain (dBGain).

        """
        self.checkExistingFilters(filterName)

        Fs = self.f_axis[-1] * 4 #2
        wc = 2 * pi * fc / Fs
        alpha = sin(wc) / (2 * Q)

        # thks T
        A = 10 ** (dBGain / 20)
        b = np.zeros(3)
        a = np.zeros(3)
        b[0] = A * (1 - cos(wc)) / 2
        b[1] = A * (1 - cos(wc))
        b[2] = A * (1 - cos(wc)) / 2
        a[0] = 1 + alpha
        a[1] = -2 * cos(wc)
        a[2] = 1 - alpha
        _, h = freqz(b/a[0], a/a[0], worN=self.f_axis, fs=Fs)

        # update crossover
        self.filterList.append(filterName)
        self.frf_List.append(h)
        self.h = updateCrossovers(self.f_axis, self.filterList, self.frf_List)
        # print(b)
        # print(a)

        out = []
        if coeff is True:
            out.append(b)
            out.append(a)
        elif coeff is False:
            out = None
        return out

    def addHighPassBQ(self, filterName, fc, Q, dBGain=0, coeff=False):
        """
        Add a biquad (BQ) high-pass filter to the crossover design.

        Parameters:
            filterName (str): Name of the filter to be added.
            fc (float): Cutoff frequency of the filter in Hertz.
            Q (float): Quality factor of the filter.
            dBGain (float): Gain in decibels applied to the filter (default: 0).
            coeff (bool): If True, return filter coefficients (default: False).

        Returns:
            list or None: If coeff is True, returns a list containing filter coefficients [b, a].
                          Otherwise, returns None.

        Notes:
            This method adds a biquad high-pass filter to the crossover design using the given
            cutoff frequency (fc), quality factor (Q), and optional gain (dBGain).

        """
        self.checkExistingFilters(filterName)

        Fs = self.f_axis[-1] * 4 #2
        wc = 2 * pi * fc / Fs
        alpha = sin(wc) / (2 * Q)

        # thks T
        A = 10 ** (dBGain / 20)
        b = np.zeros(3)
        a = np.zeros(3)
        b[0] = A * (1 + cos(wc)) / 2
        b[1] = -A * (1 + cos(wc))
        b[2] = A * (1 + cos(wc)) / 2
        a[0] = 1 + alpha
        a[1] = -2 * cos(wc)
        a[2] = 1 - alpha
        _, h = freqz(b/a[0], a/a[0], worN=self.f_axis, fs=Fs)

        # update crossovers
        self.filterList.append(filterName)
        self.frf_List.append(h)
        self.h = updateCrossovers(self.f_axis, self.filterList, self.frf_List)

        out = []
        if coeff is True:
            out.append(b)
            out.append(a)
        elif coeff is False:
            out = None
        return out

    def addBandPassBQ(self, filterName, fc, Q, dBGain=0, coeff=False):
        """
        Add a biquad (BQ) band-pass filter to the crossover design.

        Parameters:
            filterName (str): Name of the filter to be added.
            fc (float): Cutoff frequency of the filter in Hertz.
            Q (float): Quality factor of the filter.
            dBGain (float): Gain in decibels applied to the filter (default: 0).
            coeff (bool): If True, return filter coefficients (default: False).

        Returns:
            list or None: If coeff is True, returns a list containing filter coefficients [b, a].
                          Otherwise, returns None.

        Notes:
            This method adds a biquad band-pass filter to the crossover design using the given
            cutoff frequency (fc), quality factor (Q), and optional gain (dBGain).

        """
        self.checkExistingFilters(filterName)

        Fs = self.f_axis[-1] * 4 #2
        wc = 2 * pi * fc / Fs
        alpha = sin(wc) / (2 * Q)

        # thks T
        A = 10 ** (dBGain / 20)
        b = np.zeros(3)
        a = np.zeros(3)
        b[0] = A*alpha
        b[1] = 0
        b[2] = -A*alpha
        a[0] = 1 + alpha
        a[1] = -2*cos(wc)
        a[2] = 1 - alpha
        _, h = freqz(b/a[0], a/a[0], worN=self.f_axis, fs=Fs)

        # update crossovers
        self.filterList.append(filterName)
        self.frf_List.append(h)
        self.h = updateCrossovers(self.f_axis, self.filterList, self.frf_List)

        out = []
        if coeff is True:
            out.append(b)
            out.append(a)
        elif coeff is False:
            out = None
        return out

    def addPeakEQ(self, filterName, fc, Q, dBGain=0, coeff=False):
        """
        Add a biquad (BQ) peak-eq filter to the crossover design.

        Parameters:
            filterName (str): Name of the filter to be added.
            fc (float): Cutoff frequency of the filter in Hertz.
            Q (float): Quality factor of the filter.
            dBGain (float): Gain in decibels applied to the filter (default: 0).
            coeff (bool): If True, return filter coefficients (default: False).

        Returns:
            list or None: If coeff is True, returns a list containing filter coefficients [b, a].
                          Otherwise, returns None.

        Notes:
            This method adds a biquad peak-eq filter to the crossover design using the given
            cutoff frequency (fc), quality factor (Q), and optional gain (dBGain).

        """
        self.checkExistingFilters(filterName)

        Fs = self.f_axis[-1] * 4 #2
        wc = 2 * pi * fc / Fs
        alpha = sin(wc) / (2 * Q)

        # thks T
        A = 10 ** (dBGain / 40)
        b = np.zeros(3)
        a = np.zeros(3)
        b[0] = 1 + alpha * A
        b[1] = -2 * cos(wc)
        b[2] = 1 - alpha * A
        a[0] = 1 + alpha / A
        a[1] = -2 * cos(wc)
        a[2] = 1 - alpha / A
        _, h = freqz(b/a[0], a/a[0], worN=self.f_axis, fs=Fs)

        # update crossovers
        self.filterList.append(filterName)
        self.frf_List.append(h)
        self.h = updateCrossovers(self.f_axis, self.filterList, self.frf_List)

        out = []
        if coeff is True:
            out.append(b)
            out.append(a)
        elif coeff is False:
            out = None
        return out

    def addLowShelf(self, filterName, fc, Q, dBGain=0, coeff=False):
        """
        Add a biquad (BQ) low-shelf filter to the crossover design.

        Parameters:
            filterName (str): Name of the filter to be added.
            fc (float): Cutoff frequency of the filter in Hertz.
            Q (float): Quality factor of the filter.
            dBGain (float): Gain in decibels applied to the filter (default: 0).
            coeff (bool): If True, return filter coefficients (default: False).

        Returns:
            list or None: If coeff is True, returns a list containing filter coefficients [b, a].
                          Otherwise, returns None.

        Notes:
            This method adds a biquad low-shelf filter to the crossover design using the given
            cutoff frequency (fc), quality factor (Q), and optional gain (dBGain).

        """
        self.checkExistingFilters(filterName)

        Fs = self.f_axis[-1] * 4 #2
        wc = 2 * pi * fc / Fs
        alpha = sin(wc) / (2 * Q)

        # thks T
        A = 10 ** (dBGain / 40)
        b = np.zeros(3)
        a = np.zeros(3)
        b[0] = A*((A+1) - (A-1)*cos(wc) + 2*sqrt(A)*alpha)
        b[1] = 2*A*((A-1) - (A+1)*cos(wc))
        b[2] = A*((A+1) - (A-1)*cos(wc) - 2*sqrt(A)*alpha)
        a[0] = (A+1) + (A-1)*cos(wc) + 2*sqrt(A)*alpha
        a[1] = -2*((A-1) + (A+1)*cos(wc))
        a[2] = (A+1) + (A-1)*cos(wc) - 2*sqrt(A)*alpha
        _, h = freqz(b/a[0], a/a[0], worN=self.f_axis, fs=Fs)

        # update crossovers
        self.filterList.append(filterName)
        self.frf_List.append(h)
        self.h = updateCrossovers(self.f_axis, self.filterList, self.frf_List)

        out = []
        if coeff is True:
            out.append(b)
            out.append(a)
        elif coeff is False:
            out = None
        return out

    def addHighShelf(self, filterName, fc, Q, dBGain=0, coeff=False):
        """
        Add a biquad (BQ) high-shelf filter to the crossover design.

        Parameters:
            filterName (str): Name of the filter to be added.
            fc (float): Cutoff frequency of the filter in Hertz.
            Q (float): Quality factor of the filter.
            dBGain (float): Gain in decibels applied to the filter (default: 0).
            coeff (bool): If True, return filter coefficients (default: False).

        Returns:
            list or None: If coeff is True, returns a list containing filter coefficients [b, a].
                          Otherwise, returns None.

        Notes:
            This method adds a biquad high-shelf filter to the crossover design using the given
            cutoff frequency (fc), quality factor (Q), and optional gain (dBGain).

        """
        self.checkExistingFilters(filterName)

        Fs = self.f_axis[-1] * 4 #2
        wc = 2 * pi * fc / Fs
        alpha = sin(wc) / (2 * Q)

        # thks T
        A = 10 ** (dBGain / 40)
        b = np.zeros(3)
        a = np.zeros(3)
        b[0] = A*( (A+1) + (A-1)*cos(wc) + 2*sqrt(A)*alpha )
        b[1] = -2*A*( (A-1) + (A+1)*cos(wc) )
        b[2] = A*( (A+1) + (A-1)*cos(wc) - 2*sqrt(A)*alpha )
        a[0] = (A+1) - (A-1)*cos(wc) + 2*sqrt(A)*alpha
        a[1] = 2*( (A-1) - (A+1)*cos(wc) )
        a[2] = (A+1) - (A-1)*cos(wc) - 2*sqrt(A)*alpha
        _, h = freqz(b/a[0], a/a[0], worN=self.f_axis, fs=Fs)

        # update crossovers
        self.filterList.append(filterName)
        self.frf_List.append(h)
        self.h = updateCrossovers(self.f_axis, self.filterList, self.frf_List)

        out = []
        if coeff is True:
            out.append(b)
            out.append(a)
        elif coeff is False:
            out = None
        return out

    def addGain(self, filterName, dBGain):
        """
        Add a dB gain to the crossover design.

        Parameters:
            filterName (str): Name of the filter to be added.
            dBGain (float): Gain in decibels applied to the filter (default: 0).

        Returns:
            list or None: If coeff is True, returns a list containing filter coefficients [b, a].
                          Otherwise, returns None.

        Notes:
            This method adds a general gain to the crossover design using the given
            the gain (dBGain).

        """
        self.checkExistingFilters(filterName)

        h = np.ones(len(self.f_axis), dtype=complex) * 10**(dBGain/20)

        # update crossovers
        self.filterList.append(filterName)
        self.frf_List.append(h)
        self.h = updateCrossovers(self.f_axis, self.filterList, self.frf_List)
        return None

    def addTransferFunction(self, filterName, tf):
        self.checkExistingFilters(filterName)

        self.filterList.append(filterName)
        self.frf_List.append(tf)
        self.h = updateCrossovers(self.f_axis, self.filterList, self.frf_List)
        return None

    def addDelay(self, filterName, dt, timeConvention='-jwt'):
        """
        Add a time delay filter to the crossover design.

        Parameters:
            filterName (str): Name of the filter to be added.
            dt (float): time delay in seconds.
            timeConvention (str): time convention to be used (-jwt; +jwt)

        Returns:
            list or None: If coeff is True, returns a list containing filter coefficients [b, a].
                          Otherwise, returns None.

        Notes:
            This method adds a time delay to the crossover design using the given
            delay time (dt) and its time convention (timeConvention).

        """
        self.checkExistingFilters(filterName)

        if timeConvention == '-jwt' or timeConvention == '-':
            tc = -1
        elif timeConvention == '+jwt' or timeConvention == 'jwt' or timeConvention == '+':
            tc = 1
        om = 2*pi*self.f_axis
        h = np.exp(tc * 1j*om*dt)

        # update crossovers
        self.filterList.append(filterName)
        self.frf_List.append(h)
        self.h = updateCrossovers(self.f_axis, self.filterList, self.frf_List)
        return None

    def addPhaseFlip(self, filterName):
        self.checkExistingFilters(filterName)
        h = np.ones(len(self.f_axis), dtype=complex) * -1

        self.filterList.append(filterName)
        self.frf_List.append(h)
        self.h = updateCrossovers(self.f_axis, self.filterList, self.frf_List)
        return None

    def deleteFilter(self, filterName):
        # delete corresponding filter stage from crossover
        if filterName in self.filterList:
            index = self.filterList.index(filterName)
            self.filterList.pop(index)
            self.frf_List.pop(index)

            # update crossovers
        self.h = updateCrossovers(self.f_axis, self.filterList, self.frf_List)
        return None

    def checkExistingFilters(self, filterName):
        if filterName in self.filterList:
            self.deleteFilter(filterName)
        return None

    def plot(self, divide_stages=True, amplitude=64):
        # plot total response

        # get max value for ylim
        maxList = np.zeros(len(self.filterList) + 1)
        for i in range(len(self.filterList)):
            maxList[i] = np.max(gtb.gain.dB(self.frf_List[i])) + 3
        maxList[-1] = np.max(gtb.gain.dB(self.h)) + 3
        max_dB = +amplitude / 2#np.max(maxList)
        min_dB = -amplitude / 2 #max_dB - amplitude

        fig, ax = plt.subplots(2, 1)
        if divide_stages is True:
            for i in range(len(self.filterList)):
                ax[0].semilogx(self.f_axis, gtb.gain.dB(self.frf_List[i]), label=self.filterList[i])
                ax[1].semilogx(self.f_axis, np.angle(self.frf_List[i]), '--', label=self.filterList[i])
            ax[0].semilogx(self.f_axis, gtb.gain.dB(self.h), 'k', label='H(f)')
            ax[1].semilogx(self.f_axis, np.angle(self.h), '--k', label='H(f)')


        elif divide_stages is False:
            ax[0].semilogx(self.f_axis, gtb.gain.dB(self.h), 'k', label='H(f)')
            ax[1].semilogx(self.f_axis, np.angle(self.h), '--k', label='H(f)')

        else:
            Exception('divide_stages must be either True or False.')

        for i in range(2):
            ax[i].legend()
            ax[i].grid(which='both')
        ax[0].set(ylabel='Gain [dB]', ylim=[min_dB, max_dB])
        ax[1].set(xlabel='Frequency [Hz]', ylabel='Phase [rad]')
        plt.tight_layout()
        plt.show()
        return None

    def save(self, fileName):
        np.savez(fileName,
                 identifier = self.identifier,
                 f_axis     = self.f_axis,
                 h          = self.h,
                 filterList = self.filterList,
                 frf_list   = self.frf_List,
                 )
        return None


# update crossover class
def updateCrossovers(f_axis, filterList, frfList):
    # update frequency response of crossover
    nFilters = len(filterList)
    h = np.ones(len(f_axis), dtype=complex)
    for i in range(nFilters):
        h *= frfList[i]
    return h


# ACOUSTIC RESPONSE EQUALIZER
class response_eq:
    def __init__(self, freq, H, bounds):
        self.freq = freq
        self.H = np.abs(H)
        self.bounds = bounds
        self.fmin = bounds[0]
        self.fmax = bounds[1]
        self.ifmin, _ = gtb.findInArray(freq, self.fmin)
        self.ifmax, _ = gtb.findInArray(freq, self.fmax)
        self.target = np.ones(len(freq))
        self.xover = xovers(freq)

        # initialize some values
        self.intersec = None
        self.peaks = None


        # run analyser
        self.analyse_response()

    def analyse_response(self):
        self.intersec = self.find_intersections()
        self.fLogMean = self.find_logMean()
        self.peaks = self.find_peaks()
        for i in range(len(self.intersec)):
            try:
                p_opt, _ = sp.optimize.curve_fit(filter_optimize, self.freq[self.intersec[i]:self.intersec[i+1]],
                                                 - gtb.gain.dB(self.H[self.intersec[i]:self.intersec[i+1]]),
                                                 p0=[self.fLogMean[i], 1.5, gtb.gain.dB(1 + (1-self.peaks[i]))],
                                                 maxfev=10000)
                fc = p_opt[0]
                q = p_opt[1]
                g = p_opt[2]
                self.xover.addPeakEQ("peq_"+str(i), fc, q, g)
                self.H *= self.xover.frf_List[i]
            except:
                None
        return None

    def find_intersections(self):
        target_function = self.target
        transfer_function = self.H
        intersections = []
        for i in np.arange(self.ifmin, self.ifmax, 1): #range(len(target_function) - 1):
            if (target_function[i] - transfer_function[i]) * (target_function[i+1] - transfer_function[i+1]) < 0:
                intersections.append(i)
        return intersections

    def find_peaks(self):
        peaks = np.zeros(len(self.intersec) - 1)
        for i in range(len(peaks)):
            if np.mean(self.H[self.intersec[i]:self.intersec[i+1]]) < 1:
                peaks[i] = np.min(np.abs(self.H[self.intersec[i]:self.intersec[i+1]]))
            elif np.mean(self.H[self.intersec[i]:self.intersec[i+1]]) > 1:
                peaks[i] = np.max(np.abs(self.H[self.intersec[i]:self.intersec[i+1]]))
        return peaks

    def find_logMean(self):
        fLogMean = np.zeros(len(self.intersec) - 1)
        for i in range(len(fLogMean)):
            fLogMean[i] = np.sqrt(self.freq[self.intersec[i]] * self.freq[self.intersec[i+1]])
        return fLogMean


def filter_optimize(freq, fc, q, g):
    """
    Build a simple peaking EQ to fit data using scipy.optimize.curve_fit()
    """
    xo = xovers(freq)
    xo.addPeakEQ('peq', fc, q, g)
    return gtb.gain.dB(xo.h)



"""Generic access to spaces."""

import numpy as _np
import numba as _numba


def function_space(grid, kind, degree, scatter=True, **kwargs):
    """
    Initialize a function space.

    Parameters
    ----------
    grid : bempp_cl.Grid
        The grid that the space is defined on.
    kind : str
        The space type
    degree : int
        The polynomial degree of the space
    support_elements : np.array
        The element indices of elements that make up the part
        of the mesh on which the space is defined.
    segments : list
        The segment numbers of the part of the mesh on which the space is defined.
    swapped_normals : bool
        TODO
    scatter : bool
        TODO
    include_boundary_dofs : bool
        Should degrees of freedom on the boundary of the grid segments be included?
    truncate_at_segment_edge : bool
        Should basis functions be truncated at the edge of the grid segment? If this is set to true,
        continuous spaces will no longer be continuous across the segment edge.
    """
    from bempp_cl.api.utils import pool

    from . import scalar_spaces
    from . import scalar_dual_spaces
    from . import maxwell_spaces

    space_f = None

    if "support_elements" in kwargs and "segments" in kwargs:
        raise ValueError("Only one of 'support_elements' and 'segments' must be nonzero.")

    if kind == "DP":
        if degree == 0:
            space_f = scalar_spaces.p0_discontinuous_function_space
        if degree == 1:
            space_f = scalar_spaces.p1_discontinuous_function_space

    if kind == "P":
        if degree == 1:
            space_f = scalar_spaces.p1_continuous_function_space

    if kind == "DUAL":
        if degree == 0:
            space_f = scalar_dual_spaces.dual0_function_space
        if degree == 1:
            space_f = scalar_dual_spaces.dual1_function_space

    if kind == "RWG" or kind == "RT":
        if degree == 0:
            space_f = maxwell_spaces.rwg0_function_space

    if kind == "SNC" or kind == "NC":
        if degree == 0:
            space_f = maxwell_spaces.snc0_function_space

    if kind == "BC":
        if degree == 0:
            space_f = maxwell_spaces.bc_function_space

    if kind == "RBC":
        if degree == 0:
            space_f = maxwell_spaces.rbc_function_space

    if space_f is None:
        raise ValueError("Requested space not implemented.")

    space = space_f(grid, **kwargs)

    if scatter and pool.is_initialised() and not pool.is_worker():
        pool.execute(
            _space_scatter_worker,
            grid.id,
            space.id,
            kind,
            degree,
            kwargs,
        )
        space._is_scattered = True

    return space


class SpaceBuilder(object):
    """Configure and builds a space object."""

    def __init__(self, grid):
        """Set all parameters to None."""
        self._grid = grid

        self._codomain_dimension = None
        self._order = None
        self._shapeset = None
        self._local2global_map = None
        self._local_multipliers = None
        self._identifier = None
        self._support = None
        self._normal_multipliers = None
        self._barycentric_representation = None
        self._numba_evaluator = None
        self._numba_surface_gradient = None
        self._numba_surface_curl = None
        self._dof_transformation = None
        self._collocation_points = None
        self._is_localised = None
        self._global2local_map = None

        self._is_barycentric = False
        self._space = None
        self._requires_dof_transformation = False

    def set_codomain_dimension(self, codomain_dimension):
        """Set the codomain dimension."""
        self._codomain_dimension = codomain_dimension
        return self

    def set_support(self, support):
        """Set the support."""
        self._support = support
        return self

    def set_normal_multipliers(self, normal_multipliers):
        """Set normal multipliers."""
        self._normal_multipliers = normal_multipliers
        return self

    def set_order(self, order):
        """Set the order of the space."""
        self._order = order
        return self

    def set_shapeset(self, shapeset):
        """Set the shapeset string."""
        self._shapeset = shapeset
        return self

    def set_local2global(self, local2global):
        """Set the local2global map."""
        self._local2global_map = local2global
        return self

    def set_global2local(self, global2local):
        """Set the global2local map."""
        self._global2local_map = global2local
        return self

    def set_local_multipliers(self, local_multipliers):
        """Set the local multipliers."""
        self._local_multipliers = local_multipliers
        return self

    def set_identifier(self, identifier):
        """Set the identifier."""
        self._identifier = identifier
        return self

    def set_is_localised(self, is_localised):
        """Set to true if space is localised."""
        self._is_localised = is_localised
        return self

    def set_dof_transformation(self, dof_transformation):
        """Set the dof transformation."""
        self._dof_transformation = dof_transformation
        self._requires_dof_transformation = True

        return self

    def set_is_barycentric(self, is_barycentric):
        """Call to define space as barycentric."""
        self._is_barycentric = is_barycentric
        return self

    def set_numba_evaluator(self, basis_evaluator):
        """Hand over Numba method that evaluates the basis."""
        self._numba_evaluator = basis_evaluator
        return self

    def set_numba_surface_gradient(self, surface_gradient):
        """Hand over Numba method that evaluates surface gradient."""
        self._numba_surface_gradient = surface_gradient
        return self

    def set_numba_surface_curl(self, surface_curl):
        """Hand over Numba method that evaluates surface curl."""
        self._numba_surface_curl = surface_curl
        return self

    def set_barycentric_representation(self, barycentric_representation):
        """Set barycentric representation."""
        self._barycentric_representation = barycentric_representation
        return self

    def set_collocation_points(self, collocation_points):
        """Define the collocation points."""
        self._collocation_points = collocation_points
        return self

    def build(self):
        """Build a space object."""
        from scipy.sparse import identity

        if self._space is not None:
            return self._space

        # Check if enough information was provided.
        if self._codomain_dimension is None:
            raise ValueError("Codomain dimension not defined.")

        if self._order is None:
            raise ValueError("order not defined.")

        if self._shapeset is None:
            raise ValueError("shapeset not defined.")

        if self._local2global_map is None:
            raise ValueError("local2global map not defined.")

        if self._local_multipliers is None:
            raise ValueError("Local multipliers not defined.")

        if self._identifier is None:
            raise ValueError("identifier not defined.")

        if self._is_localised is None:
            raise ValueError("is_localised property not defined.")

        self._space = FunctionSpace.__new__(FunctionSpace)

        if self._support is None:
            self._support = _np.ones(self._grid.number_of_elements, dtype=_np.bool)

        if self._normal_multipliers is None:
            self._normal_multipliers = _np.ones(self._grid.number_of_elements, dtype=_np.int32)

        if self._dof_transformation is None:
            ndofs = 1 + _np.max(self._local2global_map)
            self._dof_transformation = identity(ndofs, dtype=_np.float64)

        if self._numba_evaluator is None:
            self._numba_evaluator = _numba_evaluate

        if self._barycentric_representation is None and self._is_barycentric:
            self._barycentric_representation = self._space

        self._space.__init__(
            self._grid,
            self._codomain_dimension,
            self._order,
            self._shapeset,
            self._local2global_map,
            self._global2local_map,
            self._local_multipliers,
            self._identifier,
            self._is_localised,
            self._support,
            self._normal_multipliers,
            self._requires_dof_transformation,
            self._is_barycentric,
            self._barycentric_representation,
            self._dof_transformation,
            self._numba_evaluator,
            self._numba_surface_gradient,
            self._numba_surface_curl,
            self._collocation_points,
        )

        return self._space


class FunctionSpace(object):
    """
    Main class for function spaces.

    This class is not meant to be initialized on its
    own. Rather there are functions for each type
    of space that configure this object through a
    builder pattern.

    """

    def __init__(
        self,
        grid,
        codomain_dimension,
        order,
        shapeset,
        local2global_map,
        global2local_map,
        local_multipliers,
        identifier,
        is_localised,
        support,
        normal_multipliers,
        requires_dof_transformation,
        is_barycentric,
        barycentric_representation,
        dof_transformation,
        numba_evaluator,
        numba_surface_gradient,
        numba_surface_curl,
        collocation_points,
    ):
        """Initialize the space."""
        from .shapesets import Shapeset
        from scipy.sparse import coo_matrix
        from bempp_cl.api.utils.helpers import create_unique_id

        self._grid = grid
        self._grid_id = self._grid.id
        self._codomain_dimension = codomain_dimension
        self._order = order
        self._shapeset = Shapeset(shapeset)
        self._local2global_map = local2global_map
        self._global2local_map = global2local_map
        self._local_multipliers = local_multipliers
        self._identifier = identifier
        self._support = support
        self._requires_dof_transformation = requires_dof_transformation
        self._is_barycentric = is_barycentric
        self._barycentric_representation = barycentric_representation
        self._dof_transformation = dof_transformation
        self._numba_evaluate = numba_evaluator
        self._numba_surface_gradient = numba_surface_gradient
        self._numba_surface_curl = numba_surface_curl
        self._normal_multipliers = normal_multipliers
        self._number_of_support_elements = _np.count_nonzero(self._support)
        self._support_elements = _np.flatnonzero(self._support).astype("uint32")
        self._collocation_points = collocation_points

        self._id = create_unique_id()
        self._hash_string = None
        self._color_map = None
        self._mass_matrix = None
        self._inverse_mass_matrix = None
        self._sorted_indices = None
        self._indexptr = None
        self._is_scattered = False

        # Number of dofs for the space defined over the grid
        # This is different from the global_dof_count, which
        # takes the dof transformation matrix into account.

        number_of_grid_dofs = 1 + _np.max(self._local2global_map)
        self._grid_dof_count = number_of_grid_dofs

        # Number of shape functions
        nshape_fun = self.number_of_shape_functions

        # Generate map to localised space

        self._map_to_localised_space = coo_matrix(
            (
                self._local_multipliers[self._support].ravel(),
                (
                    _np.arange(nshape_fun * self._number_of_support_elements),
                    self._local2global_map[self._support].ravel(),
                ),
            ),
            shape=(nshape_fun * self._number_of_support_elements, self._grid_dof_count),
            dtype="float64",
        ).tocsr()

        # Generate map to full grid
        # This works like the map to the localised grid.
        # But if the space is defined over a subgrid it
        # maps to the localised space on the full grid.

        self._map_to_full_grid = coo_matrix(
            (
                self._local_multipliers[self._support].ravel(),
                (
                    nshape_fun * _np.repeat(self._support_elements, nshape_fun)
                    + _np.tile(_np.arange(nshape_fun), self._number_of_support_elements),
                    self._local2global_map[self._support].ravel(),
                ),
            ),
            shape=(nshape_fun * self._grid.number_of_elements, number_of_grid_dofs),
            dtype="float64",
        ).tocsr()

        # Create localised space.

        # First check if space is already a localised space.
        # For this we need that all multipliers are 1 and all global
        # dofs are each only associated with one local dof.

        if is_localised:
            self._localised_space = self
        else:
            # Create a new localised space
            self._localised_space = make_localised_space(self)

    @property
    def grid(self):
        """Return the grid."""
        return self._grid

    @property
    def grid_id(self):
        """Return id of base grid."""
        return self._grid_id

    @property
    def codomain_dimension(self):
        """Return the codomain dimension."""
        return self._codomain_dimension

    @property
    def global_dof_count(self):
        """Return the global dof count."""
        return self._dof_transformation.shape[1]

    @property
    def grid_dof_count(self):
        """Return the grid dof count."""
        return self._grid_dof_count

    @property
    def order(self):
        """Return order of the space."""
        return self._order

    def cell_dofs(self, cell_index):
        """Return the DOF numbers associated with the cell."""
        return [
            None if _np.isclose(j, 0) else i
            for i, j in zip(self.local2global[cell_index], self.local_multipliers[cell_index])
        ]

    @property
    def local2global(self):
        """Return local to global map."""
        return self._local2global_map

    @property
    def local_multipliers(self):
        """Return the multipliers for each local dof."""
        return self._local_multipliers

    @property
    def normal_multipliers(self):
        """Return the normal multipliers for each grid element."""
        return self._normal_multipliers

    @property
    def global2local(self):
        """Return global to local map."""
        if self._global2local_map is None:
            self._global2local_map = invert_local2global(self.local2global, self.local_multipliers)
        return self._global2local_map

    @property
    def number_of_shape_functions(self):
        """Return the number of shape functions on each element."""
        return self._shapeset.number_of_shape_functions

    @property
    def number_of_support_elements(self):
        """Return the number of elements that form the support."""
        return self._number_of_support_elements

    @property
    def support_elements(self):
        """Return the list of elements on which space is supported."""
        return self._support_elements

    @property
    def identifier(self):
        """Return the identifier."""
        return self._identifier

    @property
    def support(self):
        """Return support of the space."""
        return self._support

    @property
    def id(self):
        """Return id string of the space."""
        return self._id

    @property
    def shapeset(self):
        """Return the shapeset."""
        return self._shapeset

    @property
    def localised_space(self):
        """Return the elementwise defined space."""
        return self._localised_space

    @property
    def is_localised(self):
        """Return true if space is localised."""
        return self.localised_space == self

    @property
    def color_map(self):
        """
        Return a coloring of the grid associated with the space.

        The coloring is defined such that if two elements have
        the same color then their associated global degrees of
        freedom do not intersect. This is important for the
        efficient summation of global dofs during the assembly.
        """
        if self._color_map is None:
            self._compute_color_map()
        return self._color_map

    @property
    def map_to_localised_space(self):
        """Return a sparse matrix that maps dofs to localised space."""
        return self._map_to_localised_space

    @property
    def map_to_full_grid(self):
        """Return a sparse matrix that maps dofs to localised space on full grid."""
        return self._map_to_full_grid

    @property
    def dof_transformation(self):
        """Return transformation from global dofs to space dofs."""
        return self._dof_transformation

    @property
    def requires_dof_transformation(self):
        """Return true if the dof transformation matrix is not the identity."""
        return self._requires_dof_transformation

    @property
    def is_barycentric(self):
        """Return true if space is defined over barycentric grid."""
        return self._is_barycentric

    @property
    def hash(self):
        """Return hash string for space comparison."""
        if self._hash_string is None:
            self._hash_string = self._generate_hash()

        return self._hash_string

    @property
    def numba_evaluate(self):
        """Return the basis evaluator."""
        return self._numba_evaluate

    @property
    def numba_surface_gradient(self):
        """Return the surface gradient evaluator."""
        if self._numba_surface_gradient is None:
            raise ValueError("No surface gradient define for this space.")
        return self._numba_surface_gradient

    @property
    def numba_surface_curl(self):
        """Return the surface curl evaluator."""
        if self._numba_surface_curl is None:
            raise ValueError("No surface curl define for this space.")
        return self._numba_surface_curl

    @property
    def has_surface_gradient(self):
        """Return True if surface gradient is defined."""
        return self._numba_surface_gradient is not None

    @property
    def has_surface_curl(self):
        """Return True if surface curl is defined."""
        return self._numba_surface_curl is not None

    @property
    def collocation_points(self):
        """Return collocation points."""
        return self._collocation_points

    def barycentric_representation(self):
        """Return barycentric_representation if it exists."""
        import collections

        # Lazy evaluation. By default only a function object is created.
        # On first call this is executed to produce the space.
        if isinstance(self._barycentric_representation, collections.abc.Callable):
            self._barycentric_representation = self._barycentric_representation(self)
        return self._barycentric_representation

    def map_to_points(self, quadrature_order=None, return_transpose=False):
        """
        Return a map from function space coefficients to point evaluations.

        Creates a mapping from function space coefficients to Green's fct.
        coefficients. Needed mainly for FMM evaluations. The point definition
        is the quadrature order of the underlying quadrature rule. If
        'return_transpose' is true then then transpose of the operator is returned.
        """
        return map_space_to_points(self, quadrature_order=quadrature_order, return_transpose=return_transpose)

    def get_elements_by_color(self):
        """
        Return color sorted elements and their index positions.

        This method returns a tuple (sorted_indices, indexptr) so that
        all element indices with color i are contained in
        sorted_indices[indexptr[i]:indexptr[i+1]].

        """
        if self._sorted_indices is None:
            self._sort_elements_by_color()
        return (self._sorted_indices, self._indexptr)

    def evaluate(self, element_index, local_coordinates):
        """
        Evaluate the basis on an element.

        Returns an array of the form
        (codomain_dimension, number_of_shape_functions, number_of_eval_points)
        that contains the basis functions evaluated at the given points.
        """
        return self.numba_evaluate(
            element_index,
            self.shapeset.evaluate,
            local_coordinates,
            self.grid.data(),
            self.local_multipliers,
            self.normal_multipliers,
        )

    def surface_gradient(self, element_index, local_coordinates):
        """Return the surface gradient."""
        return self.numba_surface_gradient(
            element_index,
            self.shapeset.gradient,
            local_coordinates,
            self.grid.data(),
            self.local_multipliers,
            self.normal_multipliers,
        )

    def surface_curl(self, element_index, element, local_coordinates):
        """Return the surface gradient."""
        return self.numba_surface_curl(
            element_index, element, self.shapeset.gradient, local_coordinates, self.grid.data()
        )

    def mass_matrix(self):
        """Return the mass matrix associated with this space."""
        if self._mass_matrix is None:
            from bempp_cl.api.operators.boundary.sparse import identity

            self._mass_matrix = identity(self, self, self).weak_form()

        return self._mass_matrix

    def inverse_mass_matrix(self):
        """Return the inverse mass matrix for this space."""
        from bempp_cl.api.assembly.discrete_boundary_operator import (
            InverseSparseDiscreteBoundaryOperator,
        )

        if self._inverse_mass_matrix is None:
            self._inverse_mass_matrix = InverseSparseDiscreteBoundaryOperator(self.mass_matrix())
        return self._inverse_mass_matrix

    def _generate_hash(self):
        """Generate a hash for the space object."""
        from hashlib import md5

        dof_transform_csr = self.dof_transformation.tocsr().sorted_indices()

        md5_gen = md5()

        md5_gen.update(self.local2global.tobytes())
        md5_gen.update(self.support_elements.tobytes())
        md5_gen.update(self.normal_multipliers.tobytes())
        md5_gen.update(self.local_multipliers.tobytes())
        md5_gen.update(dof_transform_csr.indices)
        md5_gen.update(dof_transform_csr.indptr)
        md5_gen.update(dof_transform_csr.data)

        return self.identifier + "_" + self.grid_id + "_" + md5_gen.hexdigest()

    def is_compatible(self, other):
        """Check if space is compatible with other space."""
        return self == other

    def _set_id(self, new_id):
        """Assign a new id string to the space."""
        self._id = new_id

    def _compute_color_map(self):
        """Compute the color map."""
        self._color_map = -_np.ones(self.grid.number_of_elements, dtype=_np.int32)
        for element_index in self.support_elements:
            neighbors = {element_index}
            global_dofs = self.local2global[element_index]
            for dof in global_dofs:
                for elem, _ in self.global2local[dof]:
                    neighbors.add(elem)
            neighbors.remove(element_index)

            neighbor_colors = self._color_map[list(neighbors)]
            self._color_map[element_index] = next(
                color for color in range(self.number_of_support_elements) if color not in neighbor_colors
            )

    def _sort_elements_by_color(self):
        """Implement elements by color computation."""
        sorted_indices = _np.empty(self.number_of_support_elements, dtype="uint32")
        ncolors = 1 + max(self.color_map)
        indexptr = _np.zeros(1 + ncolors, dtype="uint32")

        count = 0
        for index, color in enumerate(_np.arange(ncolors, dtype="uint32")):
            colors = _np.where(self.color_map == color)[0]
            colors_length = len(colors)
            sorted_indices[count : count + colors_length] = colors
            count += colors_length
            indexptr[index + 1] = count
        self._sorted_indices, self._indexptr = sorted_indices, indexptr

    def __eq__(self, other):
        """Check if spaces are compatible."""
        return check_if_compatible(self, other)


def return_compatible_representation(*args):
    """Return representation of spaces on same grid."""
    # Check if at least one space is barycentric.
    is_barycentric = any([space.is_barycentric for space in args])

    if not is_barycentric:
        return args
    else:
        # Convert spaces
        converted = [space.barycentric_representation() for space in args]
        if not all(converted):
            raise ValueError("Not all spaces have a valid barycentric representation.")
        return converted


def check_if_compatible(space1, space2):
    """Return true if two spaces are compatible."""
    if space1.id == space2.id:
        return True

    try:
        new_space1, new_space2 = return_compatible_representation(space1, space2)
        return new_space1.hash == new_space2.hash
    except:  # noqa: E722
        return False


def map_space_to_points(space, quadrature_order=None, return_transpose=False):
    """Return mapper from grid coeffs to point evaluations."""
    import bempp_cl.api
    from scipy.sparse import coo_matrix
    from scipy.sparse.linalg import aslinearoperator
    from bempp_cl.api.integration.triangle_gauss import rule

    grid = space.grid

    if quadrature_order is None:
        quadrature_order = bempp_cl.api.GLOBAL_PARAMETERS.quadrature.regular

    local_points, weights = rule(quadrature_order)

    number_of_local_points = local_points.shape[1]
    number_of_vertices = number_of_local_points * grid.number_of_elements

    data, global_indices, vertex_indices = map_space_to_points_impl(
        grid.data("double"),
        space.localised_space.local2global,
        space.localised_space.local_multipliers,
        space.localised_space.normal_multipliers,
        space.support_elements,
        space.numba_evaluate,
        space.shapeset.evaluate,
        local_points,
        weights,
        space.number_of_shape_functions,
    )

    if return_transpose:
        transform = coo_matrix(
            (data, (global_indices, vertex_indices)),
            shape=(space.localised_space.grid_dof_count, number_of_vertices),
        )

        return (
            aslinearoperator(space.dof_transformation.T)
            @ aslinearoperator(space.map_to_localised_space.T)
            @ aslinearoperator(transform)
        )
    else:
        transform = coo_matrix(
            (data, (vertex_indices, global_indices)),
            shape=(number_of_vertices, space.localised_space.grid_dof_count),
        )
        return (
            aslinearoperator(transform)
            @ aslinearoperator(space.map_to_localised_space)
            @ aslinearoperator(space.dof_transformation)
        )


@_numba.njit
def map_space_to_points_impl(
    grid_data,
    local2global,
    local_multipliers,
    normal_multipliers,
    support_elements,
    numba_evaluate,
    shape_fun,
    local_points,
    weights,
    number_of_shape_functions,
):
    """Run Numba accelerated computational parts for point map."""
    number_of_local_points = local_points.shape[1]
    number_of_support_elements = len(support_elements)

    nlocal = number_of_local_points * number_of_shape_functions

    data = _np.empty(nlocal * number_of_support_elements, dtype=_np.float64)
    global_indices = _np.empty(nlocal * number_of_support_elements, dtype=_np.int64)
    vertex_indices = _np.empty(nlocal * number_of_support_elements, dtype=_np.int64)

    for elem in support_elements:
        basis_values = (
            numba_evaluate(
                elem,
                shape_fun,
                local_points,
                grid_data,
                local_multipliers,
                normal_multipliers,
            )[0, :, :]
            * weights
            * grid_data.integration_elements[elem]
        )
        data[elem * nlocal : (1 + elem) * nlocal] = basis_values.ravel()
        for index in range(number_of_shape_functions):
            vertex_indices[
                elem * nlocal + index * number_of_local_points : elem * nlocal + (1 + index) * number_of_local_points
            ] = _np.arange(elem * number_of_local_points, (1 + elem) * number_of_local_points)
        global_indices[elem * nlocal : (1 + elem) * nlocal] = _np.repeat(local2global[elem, :], number_of_local_points)

    return (data, global_indices, vertex_indices)


def _process_segments(grid, support_elements, segments, swapped_normals):
    """Process information from support_elements and segments vars."""
    if support_elements is not None and segments is not None:
        raise ValueError("Only one of 'support_elements' and 'segments' must be nonzero.")

    if swapped_normals is None:
        swapped_normals = {}

    number_of_elements = grid.number_of_elements
    normal_multipliers = _np.zeros(number_of_elements, dtype=_np.int32)

    for element_index in range(number_of_elements):
        if grid.domain_indices[element_index] in swapped_normals:
            normal_multipliers[element_index] = -1
        else:
            normal_multipliers[element_index] = 1

    if support_elements is not None:
        support = _np.full(number_of_elements, False, dtype=bool)
        support[support_elements] = True
    elif segments is not None:
        support = _np.full(number_of_elements, False, dtype=bool)
        for element_index in range(number_of_elements):
            if grid.domain_indices[element_index] in segments:
                support[element_index] = True
    else:
        support = _np.full(number_of_elements, True, dtype=bool)

    return support, normal_multipliers


def invert_local2global(local2global_map, local_multipliers):
    """Obtain the global to local dof map from the local to global map."""
    global_dof_count = 1 + _np.max(local2global_map)
    number_of_elements = len(local2global_map)

    global2local_map = [[] for _ in range(global_dof_count)]

    for elem_index in range(number_of_elements):
        for local_index, dof in enumerate(local2global_map[elem_index]):
            if local_multipliers[elem_index, local_index] != 0:
                global2local_map[dof].append((elem_index, local_index))

    for index, elem in enumerate(global2local_map):
        global2local_map[index] = tuple(elem)

    return global2local_map


def make_localised_space(space):
    """Return the associated localised space."""
    number_of_elements = space.grid.number_of_elements
    support_size = space.number_of_support_elements
    local_size = space.number_of_shape_functions

    local2global_map = _np.zeros((number_of_elements, local_size), dtype="uint32")
    local2global_map[space.support] = _np.arange(local_size * support_size, dtype="uint32").reshape(
        (support_size, local_size)
    )

    local_multipliers = _np.zeros((number_of_elements, local_size), dtype="float64")
    local_multipliers[space.support] = 1

    surface_gradient = space.numba_surface_gradient if space.has_surface_gradient else None

    surface_curl = space.numba_surface_curl if space.has_surface_curl else None

    global2local_map = invert_local2global(local2global_map, local_multipliers)

    return (
        SpaceBuilder(space.grid)
        .set_codomain_dimension(space.codomain_dimension)
        .set_support(space.support)
        .set_normal_multipliers(space.normal_multipliers)
        .set_order(space.order)
        .set_shapeset(space.shapeset.identifier)
        .set_is_localised(True)
        .set_identifier(space.identifier + "_localised")
        .set_local2global(local2global_map)
        .set_global2local(global2local_map)
        .set_local_multipliers(local_multipliers)
        .set_numba_evaluator(space.numba_evaluate)
        .set_numba_surface_gradient(surface_gradient)
        .set_numba_surface_curl(surface_curl)
        .set_is_barycentric(space.is_barycentric)
        .build()
    )


def _space_scatter_worker(
    grid_id,
    space_id,
    kind,
    degree,
    kwargs,
):
    """Copy a space to a worker."""
    import bempp_cl.api
    from bempp_cl.api.utils import pool

    grid = pool.get_data(grid_id)
    space = bempp_cl.api.function_space(
        grid,
        kind,
        degree,
        **kwargs,
    )
    space._set_id(space_id)
    pool.insert_data(space_id, space)
    bempp_cl.api.log(f"Copied space {space.id} to worker {pool.get_id()}.", "debug")


@_numba.njit
def _numba_evaluate(
    element_index,
    shapeset_evaluate,
    local_coordinates,
    grid_data,
    local_multipliers,
    normal_multipliers,
):
    """Evaluate the basis on an element."""
    shapeset_values = shapeset_evaluate(local_coordinates)
    nshapeset = shapeset_values.shape[1]
    for index in range(nshapeset):
        shapeset_values[:, index, :] *= local_multipliers[element_index, index]
    return shapeset_values



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct  3 15:42:38 2023

@author: tom.munoz
"""

import bempp_cl.api
from bempp_cl.api.operators.boundary import helmholtz, sparse
from bempp_cl.api.operators.potential import helmholtz as helmholtz_potential
from bempp_cl.api.assembly.discrete_boundary_operator import DiagonalOperator
from scipy.sparse.linalg import gmres as scipy_gmres
from bempp_cl.api.linalg import gmres
import numpy as np
from tqdm import tqdm
import warnings
# from pyopencl import CompilerWarning
import electroacPy.general as gtb
from .ACSHelpers_ import getSurfaceAdmittance

warnings.filterwarnings("ignore", message="splu requires CSC matrix format")
warnings.filterwarnings("ignore", message="splu converted its input to CSC format")
# warnings.filterwarnings("ignore", category=CompilerWarning)

# bempp_cl.api.set_default_gpu_device_by_name('NVIDIA CUDA')
# bempp_cl.api.BOUNDARY_OPERATOR_DEVICE_TYPE = 'gpu'
# bempp_cl.api.POTENTIAL_OPERATOR_DEVICE_TYPE = 'gpu'
# bempp_cl.api.DEFAULT_PRECISION = 'single'

try:
    from pyopencl import CompilerWarning
    warnings.filterwarnings("ignore", category=CompilerWarning)
except:
    None


class bem:
    def __init__(self, meshPath, radiatingElement, velocity, frequency, 
                 domain="exterior", c_0=343, rho_0=1.22, **kwargs):
        """
        Create a BEM object.

        Parameters
        ----------
        meshPath : str
            Path to simulation mesh.
        radiatingElement : list of int
            Reference to BEM physical group(s) - should be surfaces.
        velocity : list of numpy array
            Velocity of radiating surfaces.
        frequency : numpy array
            Range of simulation.
        domain : str, optional
            Domain of simulation: "interior" or "exterior". 
            Will change the BEM equation. Default is "exterior".
        c_0 : float, optional
            Speed of sound in the propagation medium. Default is c=343 m/s (air).
        rho_0 : float, optional
            Medium density. Default is rho=1.22 kg/m^3 (air).
        
        **kwargs
        --------
        boundary_condition(s) : boundaryCondition object,
            Define boundary conditions on BEM mesh (infinite 
            reflection plane, absorption surface, etc.)
        direction : list directional vectors, 
            Use this kwarg to set non-normal velocity on mesh. ex: [[1, 0, 0], [1, 0, 0]]
            for two radiating surfaces with velocity toward +x.
        vibrometry_points : ndarray,
            Position of measured vibrometry points. Should be of size (Npoints, 3).                
        tol : float
            Tolerance of the GMRES solver. By default is 1E-5.

        Returns
        -------
        None.

        """
        # get main parameters
        self.radiatingElement = radiatingElement
        self.velocity = velocity
        self.frequency = frequency
        self.c_0 = c_0
        self.rho_0 = rho_0
        self.domain = domain
        self.kwargs = kwargs
        
        # check if mesh is v2
        self.meshPath = gtb.geometry.check_mesh(meshPath)
        
        # initialize possible boundary conditions
        self.impedanceSurfaceIndex = []
        self.surfaceImpedance = []
        
        # get kwargs
        self.boundary_conditions = None
        self.direction = False
        self.vibrometry_points = None
        self.tol = None
        self.parse_input()
        
        # other parameters
        self.Ns = len(self.radiatingElement)
        self.isComputed = False
        self.is2dData = checkVelocityInput(self.velocity)
        
        # initialize pressures and velocities arrays
        self.u_mesh = np.empty([len(frequency), self.Ns], dtype=object)  # separate sources
        self.p_mesh = np.empty([len(frequency), self.Ns], dtype=object)  # separate drivers
        self.u_total_mesh = np.empty([len(frequency)], dtype=object)   # summed sources
        self.p_total_mesh = np.empty([len(frequency)], dtype=object)   # summed sources
        
        # load simulation grid and mirror mesh if needed
        self.grid_sim = bempp_cl.api.import_grid(self.meshPath)
        self.grid_init = bempp_cl.api.import_grid(self.meshPath)
        self.grid_sim, self.sizeFactor = mirror_mesh(self.grid_init, 
                                                     self.boundary_conditions)
        self.vertices = np.shape(self.grid_sim.vertices)[1]

        # define space functions
        self.spaceP   = bempp_cl.api.function_space(self.grid_sim, "P", 1)
        self.identity = sparse.identity(self.spaceP, self.spaceP, self.spaceP)

        # initialize flow space
        self.spaceU_freq     = np.empty(self.Ns, dtype=object)
        self.u_callable_freq = np.empty(self.Ns, dtype=object)
        
        # create a list of all velocity spaces (corresponding to radiators) as well as correction coefficients
        self.correctionCoefficients = []
        dof = np.zeros(self.Ns)
        for i in range(self.Ns):
            spaceU = bempp_cl.api.function_space(self.grid_sim, "DP", 0,
                                              segments=[radiatingElement[i]])
            self.spaceU_freq[i] = spaceU
            dof[i] = int(spaceU.grid_dof_count)   # degree of freedom of each radiators
            if isinstance(self.direction, bool) is False:  # if direction has been defined by user
                self.correctionCoefficients.append(getCorrectionCoefficients(spaceU.support_elements,
                                                   spaceU.grid.normals, self.direction[i]))
            else:
                self.correctionCoefficients.append(1)
        
        # Assign vibrometric coefficients to corresponding radiators // Assign surface velocity if not vibration data
        self.dof = dof
        maxDOF = int(np.max(dof) + 1)
        self.coeff_radSurf = np.zeros([len(frequency), self.Ns, maxDOF], dtype=complex)
        for rs, isVib in enumerate(self.is2dData):
            if isVib is False: # assign velocity
                for f in range(len(self.frequency)): # TODO: maybe try to remove that loop
                    self.coeff_radSurf[f, rs, :int(dof[rs])] = velocity[rs][f]
            if isVib is True:  # assign velocity from vibration 
                self.coeff_radSurf[:, rs, :int(dof[rs])] = getRadiationCoefficients(self.spaceU_freq[rs].support_elements,
                                                                                    self.grid_init.centroids,
                                                                                    velocity[rs],
                                                                                    self.vibrometry_points[rs],
                                                                                    self.sizeFactor)
        
        self.admittanceCoeff = getSurfaceAdmittance(self.impedanceSurfaceIndex, self.surfaceImpedance, 
                                                    frequency, self.spaceP, self.c_0, self.rho_0)        
        # driver reference
        self.LEM_enclosures = None
        self.radiator = None

    def discard_frequency(self):
        return None
        
    def parse_input(self):
        """
        Detects and assign kwargs to BEMOBJ variables.

        Returns
        -------
        None.

        """
        if "boundary_conditions" in self.kwargs:
            self.boundary_conditions = self.kwargs["boundary_conditions"].parameters
            self.initialize_conditions()
        elif "boundary_condition" in self.kwargs: # just in case user forget "s"
            self.boundary_conditions = self.kwargs["boundary_condition"].parameters
            self.initialize_conditions()
        if "direction" in self.kwargs:
            self.direction = self.kwargs["direction"]
        if "vibrometry_points" in self.kwargs:
            self.vibrometry_points = self.kwargs["vibrometry_points"]
        if "tol" in self.kwargs:
            self.tol = self.kwargs["tol"]
        else:
            self.tol = 1e-5
            
    def initialize_conditions(self):
        """
        Assign boundary conditions to the impedanceSurfaceIndex and surfaceImpedance 
        variables.

        Returns
        -------
        None.

        """
        
        for bc in self.boundary_conditions:
            if bc not in ["x", "X", "y", "Y", "z", "Z"]:
                self.impedanceSurfaceIndex.append(self.boundary_conditions[bc]["index"])
                self.surfaceImpedance.append(self.boundary_conditions[bc]["impedance"])
            else:
                pass
    
    def solve(self):
        """
        Compute the Boundary Element Method (BEM) solution for the loudspeaker system.

        This method performs the BEM computations to determine the acoustic pressure distribution on the mesh
        due to the contribution of individual speakers. The total pressure distribution is also computed by summing
        up the contributions of all speakers.

        Returns
        -------
        None

        Notes
        -----
        This function calculates the acoustic pressure distribution on the mesh at various frequencies using the
        Boundary Element Method. It iterates through the frequencies and speakers, calculates the necessary BEM
        operators (double layer, single layer), and uses GMRES solver to solve the BEM equation for the acoustic
        pressure distribution. The results are stored in class attributes for further analysis.

        """
        
        if self.domain == "exterior":
            domain_operator = -1
        elif self.domain == "interior":
            domain_operator = +1
        
        omega = 2 * np.pi * self.frequency
        k = -omega / self.c_0

        # individual speakers
        self.u_mesh = np.empty([len(k), self.Ns], dtype=object)
        self.p_mesh = np.empty([len(k), self.Ns], dtype=object)

        # sum of all speakers
        self.u_total_mesh = np.empty([len(k)], dtype=object)
        self.p_total_mesh = np.empty([len(k)], dtype=object)

        # error
        self.error = np.zeros([len(k), self.Ns])

        print("Computing pressure on mesh")
        if self.admittanceCoeff is None:
            for i in tqdm(range(len(k))):
                # creation of the double layer
                double_layer = helmholtz.double_layer(self.spaceP, self.spaceP,
                                                      self.spaceP, k[i])
                lhs = double_layer + 0.5 * self.identity * domain_operator
                for rs in range(self.Ns):                
                    coeff_radSurf = self.coeff_radSurf[i, rs, :int(self.dof[rs])]
                    spaceU = bempp_cl.api.function_space(self.grid_sim, "DP", 0,
                                                      segments=[self.radiatingElement[rs]])
    
                    # get velocity on current radiator
                    u_total = bempp_cl.api.GridFunction(spaceU, 
                                                     coefficients=-coeff_radSurf *
                                                     self.correctionCoefficients[rs])
                    # single layer
                    single_layer = helmholtz.single_layer(spaceU,
                                                          self.spaceP, self.spaceP,
                                                          k[i])
    
                    # pressure over the whole surface of the loudspeaker (p_total)
                    rhs = 1j * omega[i] * self.rho_0 * single_layer * u_total
                    p_total, _ = gmres(lhs, rhs, tol=self.tol, 
                                       return_residuals=False)
    
                    self.p_mesh[i, rs] = p_total # individual speakers
                    self.u_mesh[i, rs] = u_total # individual speakers
            self.isComputed = True
            
        elif self.admittanceCoeff is not None:
            for i in tqdm(range(len(k))):
                # creation of the double layer
                double_layer = helmholtz.double_layer(self.spaceP, self.spaceP,
                                                      self.spaceP, k[i])
                # admittance single layer
                single_layer_Y = helmholtz.single_layer(self.spaceP, self.spaceP,
                                                        self.spaceP, k[i])
                # ABSORBING SURFACES
                Yn = self.admittanceCoeff[:, i]  # all admittance coeff at current frequency
                yn_fun = bempp_cl.api.GridFunction(self.spaceP, coefficients=Yn)  # ? doubts on its usefulness
                yn = DiagonalOperator(yn_fun.coefficients)

                # building left-hand-side term
                lhs = ((double_layer + 0.5*self.identity * domain_operator).weak_form()
                       - (1j*k[i]*single_layer_Y.weak_form()*yn))    
                
                for rs in range(self.Ns):
                    # RADIATING SURFACES
                    coeff_radSurf = self.coeff_radSurf[i, rs, :int(self.dof[rs])]

                    spaceU = bempp_cl.api.function_space(self.grid_sim, "DP", 0,
                                                      segments=[self.radiatingElement[rs]])

                    # get velocity on current radiator
                    u_total = bempp_cl.api.GridFunction(spaceU, 
                                                     coefficients=-coeff_radSurf *
                                                     self.correctionCoefficients[rs])

                    # single layer - radiating surface
                    single_layer = helmholtz.single_layer(spaceU,
                                                          self.spaceP, self.spaceP,
                                                          k[i])


                    rhs = 1j * omega[i] * self.rho_0 * single_layer * u_total
                    rhs = rhs.projections(self.spaceP)

                    # pressure over the whole surface of the loudspeaker (p_total)
                    p_total_coefficients, _ = scipy_gmres(lhs, rhs, rtol=self.tol)
                    p_total = bempp_cl.api.GridFunction(self.spaceP, coefficients=p_total_coefficients)

                    self.p_mesh[i, rs] = p_total  # individual speakers
                    self.u_mesh[i, rs] = u_total  # individual speakers
            self.isComputed = True 
        return None
        
    
    def getMicPressure(self, micPosition, individualSpeakers=False):
        """
        Get the pressure received at the considered microphones.

        Parameters
        ----------
        micPosition : numpy array
            Coordinates of the microphones (Cartesian). Shape: (nMic, 3)
        individualSpeakers : bool, optional
            If True, returns an array containing pressure received at each microphone from individual speakers.
            If False, returns the summed pressure received at each microphone from all speakers.
            Default is False.

        Returns
        -------
        pressure_mic : numpy array
            Pressure received at the specified microphones. Shape: (nFreq, nMic)

        Notes
        -----
        This function calculates the acoustic pressure received at the specified microphone positions for each frequency
        in the frequency array. The pressure is computed based on the BEM solution obtained from `computeBEM` method.
        It uses BEM operators (double layer, single layer) and their interactions with the speaker distributions to
        compute the pressure at each microphone position.

        """
        micPosition = np.array(micPosition).T
        nMic = np.shape(micPosition)[1]

        pressure_mic_array = np.zeros([len(self.frequency), nMic, self.Ns], dtype=complex)
        pressure_mic = np.zeros([len(self.frequency), nMic], dtype=complex)
        omega = 2 * np.pi * self.frequency
        k = -omega / self.c_0

        print("\n" + "Computing pressure at microphones")
        for i in tqdm(range(len(k))):  # looping through frequencies
            for rs in range(self.Ns):
                # pressure received at microphones
                pressure_mic_array[i, :, rs] = np.reshape(
                    helmholtz_potential.double_layer(self.spaceP,
                                              micPosition,
                                              k[i]) * self.p_mesh[i, rs] \
                    - 1j * omega[i] * self.rho_0 * \
                    helmholtz_potential.single_layer(
                        self.spaceU_freq[rs],
                        micPosition, k[i]) * self.u_mesh[i, rs], nMic)
                pressure_mic[i, :] += pressure_mic_array[i, :, rs]

        if individualSpeakers is True:
            out = (pressure_mic, pressure_mic_array)
        elif individualSpeakers is False:
            out = pressure_mic
        return out



# %%useful functions
def checkVelocityInput(velocity):
    """
    Check the velocity input parameter for vibrometric data: if velocity[i] is 1 dimensional: velocity,
    if velocity[i] is 2 dimensional: vibrometric data.
    :param velocity:
    :return:
    """
    isVibData      = []
    for i in range(len(velocity)):
        if len(velocity[i].shape) == 1:
            isVibData.append(False)
        elif len(velocity[i].shape) == 2:
            isVibData.append(True)
    return isVibData


def mirror_mesh(grid_init, boundary_conditions):
    bc = boundary_conditions
    size_factor = 1
    grid_tot = grid_init
    if bc is not None:
        for item in bc:
            boundary = bc[item]
            if boundary["type"] == "infinite_baffle":
                offset = boundary["offset"]
                vertices = np.copy(grid_tot.vertices)
                elements = np.copy(grid_tot.elements)
                if item in ["x", "X"]:
                    if offset != 0:
                        vertices[0, :] = 2*offset - vertices[0, :]
                        elements[[2, 0], :] = elements[[0, 2], :]
                    else:
                        vertices[0, :] = vertices[0, :]
                        elements[[2, 0], :] = elements[[0, 2], :]                    
                elif item in ["y", "Y"]:
                    if offset != 0 :
                        vertices[1, :] = 2*offset - vertices[1, :]
                        elements[[2, 0], :] = elements[[0, 2], :]
                    else:
                        vertices[1, :] = vertices[1, :]
                        elements[[2, 0], :] = elements[[0, 2], :]    
                elif item in ["z", "Z"]:
                    if offset is not False:
                        vertices[2, :] = 2*offset - vertices[2, :]
                        elements[[2, 0], :] = elements[[0, 2], :]
                    else:
                        vertices[2, :] = vertices[2, :]
                        elements[[2, 0], :] = elements[[0, 2], :]               
                else:
                    print("{} not infinite baffle.".format(item))
                grid_mirror = bempp_cl.api.Grid(vertices, elements, domain_indices=grid_tot.domain_indices)
                grid_tot = bempp_cl.api.grid.union([grid_tot, grid_mirror],
                                                [grid_tot.domain_indices, grid_mirror.domain_indices])
                size_factor *= 2
    return grid_tot, size_factor


#%% Radiation coefficients
def getRadiationCoefficients(support_elements, centroids, vibrometric_data, 
                             vibrometry_points, sizeFactor):
    """
    Build radiation coefficients from vibrometric dataset.

    :param support_elements:
    :param centroids:
    :param vibrometric_data:
    :return:
    """
    
    if sizeFactor > 1:
        slices = gtb.slice_array_into_parts(support_elements, sizeFactor)
        vertex_location = centroids[slices[0], :]          # get [x, y, z] vertex position of radiator
        vertex_center   = gtb.geometry.recenterZero(vertex_location)    # recenter geometry at [x=0, y=0, z=0]
    else:
        vertex_location = centroids[support_elements, :]  # get [x, y, z] vertex position of radiator
        vertex_center = gtb.geometry.recenterZero(vertex_location)  # recenter geometry at [x=0, y=0, z=0]

    Nfft = np.shape(vibrometric_data)[1]
    _, coefficients = gtb.geometry.points_within_radius(vertex_center,
                                                        vibrometry_points,
                                                        5e-3, vibrometric_data, Nfft)

    print("COEFFICIENTS: ", coefficients.shape)
    if sizeFactor > 1:
        coefficients = np.tile(coefficients, (1, sizeFactor))
    return coefficients

def getCorrectionCoefficients(support_elements, normals, direction):
    """
    Depending on the radiator's defined radiation direction, this function returns coefficients that are to be applied
    on the triangles of the considered radiator. The user can set specific direction on a single radiator only: by
    setting corresponding direction to False (direction will thus be normal to elementary surfaces)
    :param support_elements:
    :param normals:
    :param direction:
    :return:
    """
    correctionCoefficients = np.zeros(len(support_elements))

    if isinstance(direction, bool) is False: # check for cases like [[1, 0, 0], False, [1, 0, 0]]
        for i, element in enumerate(support_elements):
            correctionCoefficients[i] = (np.dot(direction, normals[element, :]) /
                                               (np.linalg.norm(normals[element, :]) * np.linalg.norm(direction)))
    else: # if [x, y, z] norm is not defined, set radiation to be normal to elementary surfaces
        correctionCoefficients = 1
    return correctionCoefficients


#%% Impedance functions
def get_group_points(grid, group_number):
    domain_indices = grid.domain_indices
    elements       = grid.elements
    
    # Find the indices where domain_indices match the group_number
    group_indices = np.where(domain_indices == group_number)  # indices of support elements

    # Get the corresponding columns of elements
    group_points = elements[:, group_indices]  # segments (and NOT points!) part of group_number

    # Reshape to a 1D array and use unique to get unique values
    unique_points   = np.unique(group_points)

    return unique_points, group_indices

#%% boundary conditions
class boundaryConditions:
    def __init__(self, rho=1.22, c=343):
        self.parameters = {}
        self.c = c
        self.rho = rho
        self.Zc = rho*c
    
    def addInfiniteBoundary(self, normal, offset=0, **kwargs):
        if normal not in ["x", "y", "z", "X", "Y", "Z"]:
            raise ValueError("Normal to axis should be x, y or z.")
        self.parameters[normal] = {}
        self.parameters[normal]["offset"] = offset
        self.parameters[normal]["type"] = "infinite_baffle"
        if "absorption" in kwargs:
            self.parameters[normal]["absorption"] = kwargs["absorption"]
        elif "impedance" in kwargs:
            self.parameters[normal]["impedance"] = kwargs["impedance"]
        if "frequency" in kwargs:
            self.parameters[normal]["frequency"] = kwargs["frequency"]
        else:
            None
            
    def addSurfaceImpedance(self, name, index, data_type, value,
                            frequency=None, targetFrequency=None,
                            interpolation="linear"):
        """
        Add an impedance to a surface.

        Parameters
        ----------
        name : str,
            name of surface.
        index : int
            reference to BEM mesh.
        data_type : float or array of float
            Type of data: "impedance", "absorption", "reflection", "admittance".
        value : numpy array
            Data to apply on surface.
        frequency : numpy array, optional
            Range of impedance data (if using frequency-dependant impedance).
            The default is None.
        targetFrequency : numpy array, optional
            BEM frequency range - use it if impedance data as a different frequency-axis
            than the BEM study. The default is None.
        interpolation : str, optional
            How to interpolate impedance data on frequency range. Either "linear" or "cubic".
            The default is "linear".
            
        Returns
        -------
        None.
        
        Notes
        -----
        If data_type is "impedance", the corresponding value should be given without 
        normalization to characteristic impedance of air.

        """
        
        self.parameters[name] = {}
        self.parameters[name]["index"] = index
        self.parameters[name]["data_type"] = data_type
        self.parameters[name]["type"] = "absorption_surface"
        self.parameters[name]["frequency"] = frequency
        
        if data_type == "impedance":
            self.parameters[name]["impedance"] = value
            Z = value
            self.parameters[name]["admittance"] = 1/Z
            self.parameters[name]["absorption"] = 1 - np.abs((Z-self.Zc)/(Z+self.Zc))**2
        elif data_type == "reflection":
            self.parameters[name]["reflection"] = value
            self.parameters[name]["impedance"] = (1+value) / (1-value) * self.Zc
            self.parameters[name]["admittance"] = 1/self.parameters[name]["impedance"]
        elif data_type == "absorption":
            self.parameters[name]["absorption"] = value
            self.parameters[name]["impedance"] = (2-value) / value *  \
                                                self.Zc
            self.parameters[name]["admittance"] = 1/self.parameters[name]["impedance"]
        elif data_type == "admittance":
            self.parameters[name]["admittance"] = value
            self.parameters[name]["impedance"] = 1 / value
        else:
            raise ValueError("'data_type' not understood. Try 'impedance', " +
                             "'admittance', 'reflection' or 'absorption'")
        
        if isinstance(frequency, np.ndarray) is True and isinstance(value, np.ndarray) is True:
            if len(frequency) != len(value):
                raise Exception("'value' and 'frequency' should have the same size.")
            
            if targetFrequency is None:
                self.parameters[name]["frequency"] = frequency
            else:
                if interpolation in ["linear", None]:
                    self.parameters[name]["impedance"] = np.interp(targetFrequency, 
                                                                   frequency, 
                                                                   self.parameters[name]["impedance"])
                    self.parameters[name]["admittance"] = np.interp(targetFrequency, 
                                                                    frequency, 
                                                                    self.parameters[name]["admittance"])
                    self.parameters[name]["frequency"] = targetFrequency
                elif interpolation == "cubic":
                    from scipy.interpolate import CubicSpline
                    csImp = CubicSpline(frequency, self.parameters[name]["impedance"])
                    csAdm = CubicSpline(frequency, self.parameters[name]["admittance"])
                    self.parameters[name]["impedance"] = csImp(targetFrequency)
                    self.parameters[name]["admittance"] = csAdm(targetFrequency)
                    self.parameters[name]["frequency"] = targetFrequency
                


import numpy as np
import xml.etree.ElementTree as ET
import json


# Define a function to extract desired values
def basta_lp_loader(file_path):
    # Dictionary to store the extracted values
    extracted_data = {}

    # Define the keys you want to extract with their user-friendly names
    keys_mapping = {
        "ss": "Sd",
        "cms": "Cms",
        "mms": "Mms",
        "bl": "Bl",
        "Qts": "Qts",
        "Re": "Re",
        "Rms": "Rms",
        "Qes": "Qes",
        "Le": "Le"
    }

    # Open the file and read line by line
    with open(file_path, 'r') as file:
        for line in file:
            line = line.strip()

            # Stop processing when [fgprio] section is reached
            if line.startswith('[fgprio]'):
                break

            # Split each line by '=' to extract key-value pairs
            if '=' in line:
                key, value = line.split('=')
                key = key.strip()  # Clean up key
                value = value.strip()  # Clean up value

                # Check if the key is in the mapping and not already extracted
                if key in keys_mapping:
                    extracted_data[keys_mapping[key]] = float(value)

    return extracted_data


# Define a function to extract desired values from XML
def qspeaker_lp_loader(file_path):
    # Parse the XML file
    tree = ET.parse(file_path)
    root = tree.getroot()

    # Find the speaker element
    speaker = root.find('speaker')

    # Dictionary to store the extracted values
    extracted_data = {}

    # Extract the relevant attributes from the speaker element
    if speaker is not None:
        # Define the keys you want to extract
        keys = {
            "le": "Le",
            "re": "Re",
            "bl": "Bl",
            "sd": "Sd",
            "qes": "Qes",
            "qms": "Qms",
            "qts": "Qts",
            "fs": "Fs"
                
        }
        
        # Extract the values and store in the dictionary
        for xml_key, user_friendly_key in keys.items():
            extracted_data[user_friendly_key] = float(speaker.attrib.get(xml_key, None))
    
    
    Qes = extracted_data["Qes"]
    Qms = extracted_data["Qms"]
    Bl  = extracted_data["Bl"]
    Re  = extracted_data["Re"]
    Fs  = extracted_data["Fs"]
    extracted_data["Mms"] = round(Qes * Bl**2 * 1/2/np.pi/Fs/Re, 5)
    extracted_data["Cms"] = round((1/2/np.pi/Fs)**2  / 
                                  extracted_data["Mms"], 6)
    extracted_data["Rms"] = 2*np.pi*Fs*extracted_data["Mms"]/Qms
    return extracted_data



# Define a function to extract desired values from JSON
def speakerSim_lp_loader(file_path):
    # Open and load the JSON file
    with open(file_path, 'r') as file:
        data = json.load(file)

    # Define the keys you want to extract
    keys = {
        "Le": "Le",
        "Re": "Re",
        "Bl": "Bl",
        "Sd": "Sd",
        "Cms": "Cms",
        "Mms": "Mms",
        "Rms": "Rms"
    }

    # Dictionary to store the extracted values
    extracted_data = {}

    # Extract the values and store in the dictionary
    for json_key, user_friendly_key in keys.items():
        extracted_data[user_friendly_key] = float(data.get(json_key, None))

    return extracted_data


def hornResp_lp_loader(file_path):
    # Dictionary to store the extracted values
    extracted_data = {}

    # Define the keys you want to extract
    keys = {
        "Sd": "Sd",
        "Bl": "Bl",
        "Cms": "Cms",
        "Rms": "Rms",
        "Mmd": "Mms",  # Using Mms instead of Mmd for consistency
        "Le": "Le",
        "Re": "Re"
    }

    # Open the file and read line by line
    with open(file_path, 'r') as file:
        for line in file:
            # Split each line by '='
            if '=' in line:
                key, value = line.strip().split('=')
                key = key.strip()  # Clean up key
                value = value.strip()  # Clean up value

                # Check if the key is one of the desired ones and not already extracted
                if key in keys and keys[key] not in extracted_data:
                    extracted_data[keys[key]] = float(value)

    return extracted_data


def winSd_lp_loader(file_path):
    # Dictionary to store the extracted values
    extracted_data = {}

    # Define the keys you want to extract
    keys = {
        "Le": "Le",
        "Re": "Re",
        "BL": "Bl",  # Case-insensitive key matching for Bl
        "Sd": "Sd",
        "Cms": "Cms",
        "Mms": "Mms",
        "Rms": "Rms"
    }

    # Open the file and read line by line
    with open(file_path, 'r') as file:
        for line in file:
            # Split each line by '=' and ensure valid format
            if '=' in line:
                key, value = line.strip().split('=')
                key = key.strip()  # Clean up the key
                value = value.strip()  # Clean up the value

                # Check if the key is one of the desired ones and not already extracted
                if key in keys and keys[key] not in extracted_data:
                    extracted_data[keys[key]] = float(value)

    return extracted_data


def klippel_lp_loader(file_path):
    # Dictionary to store the extracted values
    extracted_data = {}

    # Define the keys you want to extract
    keys = {
        "Re": "Re",
        "Le": "Le",
        "Bl": "Bl",
        "Sd": "Sd",
        "Cms": "Cms",
        "Mms": "Mms",
        "Rms": "Rms"
    }

    # Open the file and read line by line
    with open(file_path, 'r') as file:
        for line in file:
            # Split the line by tab or spaces if it contains a parameter
            if any(key in line for key in keys):
                parts = line.strip().split()  # Split the line by whitespace

                if len(parts) > 1:
                    key = parts[0]  # First part of the line is the key
                    value = parts[1]  # Second part of the line is the value

                    # Check if the key is one of the desired ones and not already extracted
                    if key in keys and keys[key] not in extracted_data:
                        extracted_data[keys[key]] = float(value)

    return extracted_data








"""Helper routines for Chebychev interpolation."""

import numba as _numba
import numpy as _np

# pylint: disable = C0103


class ChebychevInterpolation(object):
    """Class providing methods for Chebychev interpolation."""

    def __init__(self, order):
        self._order = order
        self._nterms = 1 + order

        self._nodes = None
        self._weights = None

        self._nodes, self._weights = chebychev_nodes_and_weights_second_kind(order)

        self._cheb_diff_mat = chebychev_differentiation_matrix(self._nodes, self._weights)

    @property
    def nodes(self):
        """Return nodes."""
        return self._nodes

    @property
    def weights(self):
        """Return weights."""
        return self._weights

    @property
    def differentiation_matrix(self):
        """Return Chebychev differentiation matrix."""
        return self._cheb_diff_mat

    def evaluate(self, values, evaluation_points):
        """
        Evaluate interpolation polynomial.

        Evaluates the interpolation polynomial with data
        given in the array values at the points in
        the array evaluation_points.
        """
        return evaluate_interp_polynomial(self.nodes, self.weights, values, evaluation_points)

    def differentiate(self, values):
        """Differentiate the polynomial defined by values."""
        return self.differentiation_matrix.dot(values)


def chebychev_nodes_and_weights_second_kind(order):
    """
    Return the Chebychev nodes and weights for a given order.

    This function computes the Chebychev nodes of the second
    kind defined as x_j = cos (pi * j / order) for j = 0 .. order.

    The corresponding weights w_j are required for the evaluation
    of the associated interpolation polynomial via barycentric
    interpolation. For the Chebychev nodes of the second kind
    it holds that w_j = (-1)^j * d_j with d_j = 1/2 for j = 0
    or j = order and d_j = 1 otherwise.
    """
    if order == 0:
        return _np.array([0.0]), _np.array([1.0])

    nodes = _np.cos(_np.pi * _np.arange(1 + order) / order)
    weights = _np.ones(1 + order, dtype="float64")
    weights[0] = weights[order] = 0.5
    weights *= _np.array([-1], dtype="float64") ** _np.arange(1 + order)

    return nodes, weights


# def evaluate_kernel_on_interpolation_points(
#    kernel_type,
#    lboundx,
#    uboundx,
#    lboundy,
#    uboundy,
#    nodes,
#    device_interface=None,
#    precision=None,
#    **kwargs
# ):
#    """
#    Return a kernel evaluation on Chebychev points.

#    This routine supports the evaluation of Laplace and
#    Helmholtz kernels of the form k(x, y) on 3d Tensor grids.
#    It returns a matrix (k(x,_i, y_j)), where the x_i
#    are points in the box defined by lboundx and uboundx and
#    y_i are points in the box defined by lboundy and uboundy.

#    The x_i and y_i are obtained from the 1d array of nodes by
#    iterating the nodes array over the three space dimensions
#    and scaling them appropriately to fit in the boxes.
#    The points are generated in the same order as the following
#    code snippet (scaling excluded for simplicity):
#
#    npoints = len(nodes)
#    tensor_points = _np.empty((npoints**3, 3), _np.float64)
#    for i in range(npoints):
#        for j in range(npoints):
#            for k in range(npoints):
#                tensor_points[i * npoints**2 + j * npoints + k, :] = \
#                    nodes[i], nodes[j], nodes[k]
#
#    Hence, the inner dimension is the fastest changing one.
#
#    Attributes
#    ----------
#    kernel_type : string
#        Either 'laplace' or 'helmholtz' to specify whether to
#        evaluate a Laplace or a Helmholtz kernel.
#    lboundx : Numpy array
#        An array specifying the lower bound for the x-component
#        box.
#    uboundx : Numpy array
#        An array specifying the upper bound for the x-component box.
#    lboundy : Numpy array
#        An array specifying the lower bound for the y-component
#        box.
#    uboundy : Numpy array
#        An array specifying the upper bound for the y-component box.
#    nodes : array
#        An array of 1d nodes defined in the interval [-1, 1] which are
#        used as basis for the tensor points.
#    kwargs : keyword arguments
#        The currently supported keyword argument is wavenumber
#        to specify the wavenumber for Helmholtz kernels.
#    """

# from bempp_cl.core import cl_helpers
# import bempp_cl.api

# nnodes = len(nodes)

# if kernel_type not in {"laplace", "helmholtz"}:
#     raise ValueError("'kernel_type' must be one of 'laplace' or 'helmholtz'.")

# if device_interface is None:
#     device_interface = bempp_cl.api.default_device()

# if precision is None:
#     precision = bempp_cl.api.get_precision(device_interface)

# real_type = cl_helpers.get_type(precision).real

# if kernel_type == "laplace":
#     dtype = cl_helpers.get_type(precision).real

# if kernel_type == "helmholtz":
#     dtype = cl_helpers.get_type(precision).complex

# result = cl_helpers.DeviceBuffer(
#     (nnodes ** 3, nnodes ** 3),
#     dtype,
#     device_interface.context,
#     access_mode="write_only",
# )

# nodes = cl_helpers.DeviceBuffer.from_array(
#     nodes, device_interface, dtype=real_type, access_mode="read_only"
# )

# options = dict()
# options["X_XMIN"] = lboundx[0]
# options["X_YMIN"] = lboundx[1]
# options["X_ZMIN"] = lboundx[2]

# options["Y_XMIN"] = lboundy[0]
# options["Y_YMIN"] = lboundy[1]
# options["Y_ZMIN"] = lboundy[2]

# options["X_XMAX"] = uboundx[0]
# options["X_YMAX"] = uboundx[1]
# options["X_ZMAX"] = uboundx[2]

# options["Y_XMAX"] = uboundy[0]
# options["Y_YMAX"] = uboundy[1]
# options["Y_ZMAX"] = uboundy[2]
# options["NNODES"] = nnodes

# if kernel_type == "helmholtz":
#     options["WAVENUMBER"] = kwargs["wavenumber"]

# cl_kernel_source = cl_helpers.kernel_source_from_identifier(
#     kernel_type + "_kernel_evaluator" + "_novec", options
# )

# cl_kernel = cl_helpers.Kernel(cl_kernel_source, device_interface.context, precision)

# event = cl_kernel.run(
#     device_interface, (nnodes ** 3, nnodes ** 3), (1, 1), nodes, result
# )
# event.wait()

# return result.get_host_copy(device_interface)


def evaluate_kernel_on_interpolation_points(
    kernel_type,
    lboundx,
    uboundx,
    lboundy,
    uboundy,
    nodes,
    device_interface=None,
    precision="double",
    **kwargs,
):
    """
    Return a kernel evaluation on Chebychev points.

    This routine supports the evaluation of Laplace and
    Helmholtz kernels of the form k(x, y) on 3d Tensor grids.
    It returns a matrix (k(x,_i, y_j)), where the x_i
    are points in the box defined by lboundx and uboundx and
    y_i are points in the box defined by lboundy and uboundy.

    The x_i and y_i are obtained from the 1d array of nodes by
    iterating the nodes array over the three space dimensions
    and scaling them appropriately to fit in the boxes.
    The points are generated in the same order as the following
    code snippet (scaling excluded for simplicity):

    npoints = len(nodes)
    tensor_points = _np.empty((npoints**3, 3), _np.float64)
    for i in range(npoints):
        for j in range(npoints):
            for k in range(npoints):
                tensor_points[i * npoints**2 + j * npoints + k, :] = \
                    nodes[i], nodes[j], nodes[k]

    Hence, the inner dimension is the fastest changing one.

    Attributes
    ----------
    kernel_type : string
        Either 'laplace' or 'helmholtz' to specify whether to
        evaluate a Laplace or a Helmholtz kernel.
    lboundx : Numpy array
        An array specifying the lower bound for the x-component
        box.
    uboundx : Numpy array
        An array specifying the upper bound for the x-component box.
    lboundy : Numpy array
        An array specifying the lower bound for the y-component
        box.
    uboundy : Numpy array
        An array specifying the upper bound for the y-component box.
    nodes : array
        An array of 1d nodes defined in the interval [-1, 1] which are
        used as basis for the tensor points.
    kwargs : keyword arguments
        The currently supported keyword argument is wavenumber
        to specify the wavenumber for Helmholtz kernels.
    """
    pointsx = chebychev_tensor_points_3d(lboundx, uboundx, nodes)
    pointsy = chebychev_tensor_points_3d(lboundy, uboundy, nodes)

    if kernel_type == "laplace":
        return evaluate_laplace_kernel_on_interpolation_points(pointsx, pointsy)

    if kernel_type == "helmholtz":
        return evaluate_helmholtz_kernel_on_interpolation_points(pointsx, pointsy, kwargs["wavenumber"])

    raise ValueError(f"Unknown kernel: {kernel_type}")


@_numba.njit(cache=True)
def evaluate_laplace_kernel_on_interpolation_points(pointsx, pointsy):
    """Evaluate the Laplace kernel at the given points."""
    values = _np.empty((pointsx.shape[0], pointsy.shape[0]), _np.float64)
    for i, x in enumerate(pointsx):
        for j, y in enumerate(pointsy):
            values[i, j] = 1 / (4 * _np.pi * _np.linalg.norm(x - y))
    return values


@_numba.njit(cache=True)
def evaluate_helmholtz_kernel_on_interpolation_points(pointsx, pointsy, wavenumber):
    """Evaluate the Laplace kernel at the given points."""
    values = _np.empty((pointsx.shape[0], pointsy.shape[0]), _np.complex128)
    for i, x in enumerate(pointsx):
        for j, y in enumerate(pointsy):
            values[i, :] = _np.exp(1j * wavenumber * _np.linalg.norm(x - y)) / (4 * _np.pi * _np.linalg.norm(x - y))
    return values


@_numba.njit(cache=True)
def chebychev_tensor_points_3d(lbound, ubound, nodes):
    """Create Chebychev points associated with a 3d tensor grid."""
    points_i = lbound[0] + 0.5 * (ubound[0] - lbound[0]) * (1 + nodes)
    points_j = lbound[1] + 0.5 * (ubound[1] - lbound[1]) * (1 + nodes)
    points_k = lbound[2] + 0.5 * (ubound[2] - lbound[2]) * (1 + nodes)

    npoints = len(nodes)
    tensor_points = _np.empty((npoints**3, 3), _np.float64)
    for i in range(npoints):
        for j in range(npoints):
            for k in range(npoints):
                tensor_points[i * npoints**2 + j * npoints + k, :] = (
                    points_i[i],
                    points_j[j],
                    points_k[k],
                )
    return tensor_points


@_numba.njit(cache=True)
def evaluate_interp_polynomial(nodes, weights, values, evaluation_points):
    """
    Evaluate an interpolation polynomial.

    This function uses barycentric evaluation for
    stability. The data values are assumed to be real.
    """
    npoints = len(evaluation_points)
    nterms = len(nodes)

    numerator = _np.zeros(npoints, _np.float64)
    denominator = _np.zeros(npoints, _np.float64)
    exact = -_np.ones(npoints, _np.int32)

    for index in range(nterms):
        xdiff = evaluation_points - nodes[index]
        temp = weights[index] / xdiff
        numerator += temp * values[index]
        denominator += temp
        exact[xdiff == 0] = index
    result = numerator / denominator
    indices = _np.where(exact > -1)[0]
    result[indices] = values[exact[indices]]
    return result


def evaluate_tensor_interp_polynomial(nodes, weights, values, evaluation_points):
    """
    Evaluate a tensor interpolation polynomial.

    This function evaluates a tensor chebychev basis with
    specified weights at a given set of evaluation points.

    Attributes
    ----------
    nodes : Numpy array
        1d array of interpolation points
    weights : Numpy array
        Associated array of barycentric weights
    values : Numpy array
        3d array of values at interpolation points.
        The dimensions specify the values along the
        (x, y, z) axis with z being the inner most axis.
        For k interpolation nodes there are k values along
        each dimension.
    evaluation_points : Numpy array
        (N x 3) array of N evaluation points in 3 dimension.
    """
    npoints = len(evaluation_points)
    nterms = len(nodes)

    output = _np.empty(npoints, _np.float64)

    for point_index, point in enumerate(evaluation_points):
        local_values = values.copy()
        for dim in range(2, -1, -1):
            numerator = _np.zeros(dim * [nterms], _np.float64)
            exact_index = -1
            denominator = 0.0
            for index in range(nterms):
                xdiff = point[dim] - nodes[index]
                if xdiff == 0.0:
                    exact_index = index
                    break
                temp = weights[index] / xdiff
                numerator += temp * local_values[..., index]
                denominator += temp
            if exact_index > -1:
                local_values = local_values[..., exact_index]
            else:
                local_values = numerator / denominator
        output[point_index] = local_values

    return output


def chebychev_differentiation_matrix(nodes, weights):
    """Return a Chebychev differentiation matrix."""
    nterms = len(nodes)
    if nterms == 1:
        return _np.array([0.0])

    node_vec = nodes.reshape(nterms, 1).copy()
    weights_vec = weights.reshape(nterms, 1).copy()
    weights_vec[0] *= 4
    weights_vec[nterms - 1] *= 4
    X = _np.tile(node_vec, (1, nterms))
    delta = X - X.T
    cheb_mat = _np.dot(weights_vec, 1.0 / weights_vec.T) / (delta + _np.eye(nterms))
    cheb_mat -= _np.diag(_np.sum(cheb_mat.T, axis=0))
    return cheb_mat
